@@ -4,7 +4,8 @@
  * Copyright (C) 2010-2011 Vic Lee
  * Copyright (C) 2011 Marc-Andre Moreau <marcandre.moreau@gmail.com>
  * Copyright (C) 2014-2015 Antenore Gatta, Fabio Castelli, Giovanni Panozzo
- * Copyright (C) 2016-2021 Antenore Gatta, Giovanni Panozzo
+ * Copyright (C) 2016-2022 Antenore Gatta, Giovanni Panozzo
+ * Copyright (C) 2022-2023 Antenore Gatta, Giovanni Panozzo, Hiroyuki Tanaka
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -36,45 +37,66 @@
  *
  */
 
-#include "rdp_plugin.h"
 #include "rdp_cliprdr.h"
 #include "rdp_event.h"
 #include "rdp_monitor.h"
 #include "rdp_settings.h"
 #include <gdk/gdkkeysyms.h>
+#ifdef GDK_WINDOWING_X11
 #include <cairo/cairo-xlib.h>
+#else
+#include <cairo/cairo.h>
+#endif
 #include <freerdp/locale/keyboard.h>
 
-static gboolean remmina_rdp_event_on_map(GtkWindow *window, GdkEvent *event, RemminaProtocolWidget *gp)
+gboolean remmina_rdp_event_on_map(RemminaProtocolWidget *gp)
 {
 	TRACE_CALL(__func__);
-	rfContext* rfi = GET_PLUGIN_DATA(gp);
-	rdpGdi* gdi;
+	rfContext *rfi = GET_PLUGIN_DATA(gp);
+	rdpGdi *gdi;
 
 	if (rfi == NULL)
 		return false;
 
-	gdi = ((rdpContext *)rfi)->gdi;
+	RemminaFile *remminafile = remmina_plugin_service->protocol_plugin_get_file(gp);
+	int do_suppress = !remmina_plugin_service->file_get_int(remminafile, "no-suppress", FALSE);
 
-	REMMINA_PLUGIN_DEBUG("Map event received, disabling TS_SUPPRESS_OUTPUT_PDU ");
-	gdi_send_suppress_output(gdi, FALSE);
+	if (do_suppress) {
+		gdi = ((rdpContext *)rfi)->gdi;
+
+		REMMINA_PLUGIN_DEBUG("Map event received, disabling TS_SUPPRESS_OUTPUT_PDU ");
+		gdi_send_suppress_output(gdi, FALSE);
+	}
 
 	return FALSE;
 }
 
-static gboolean remmina_rdp_event_on_unmap(GtkWindow *window, GdkEvent *event, RemminaProtocolWidget *gp)
+gboolean remmina_rdp_event_on_unmap(RemminaProtocolWidget *gp)
 {
 	TRACE_CALL(__func__);
-	rfContext* rfi = GET_PLUGIN_DATA(gp);
-	rdpGdi* gdi;
+	rfContext *rfi = GET_PLUGIN_DATA(gp);
+	rdpGdi *gdi;
 
 	if (rfi == NULL)
 		return false;
 
-	gdi = ((rdpContext *)rfi)->gdi;
+	GtkWidget *toplevel = gtk_widget_get_toplevel(GTK_WIDGET(gp));
+	GdkWindow *window = gtk_widget_get_window(toplevel);
 
-	REMMINA_PLUGIN_DEBUG("Unmap event received, enabling TS_SUPPRESS_OUTPUT_PDU ");
-	gdi_send_suppress_output(gdi, TRUE);
+	if (gdk_window_get_fullscreen_mode(window) == GDK_FULLSCREEN_ON_ALL_MONITORS) {
+		REMMINA_PLUGIN_DEBUG("Unmap event received, but cannot enable TS_SUPPRESS_OUTPUT_PDU when in fullscreen");
+		return FALSE;
+	}
+
+	RemminaFile *remminafile = remmina_plugin_service->protocol_plugin_get_file(gp);
+	int do_suppress = !remmina_plugin_service->file_get_int(remminafile, "no-suppress", FALSE);
+
+	if (do_suppress) {
+		gdi = ((rdpContext *)rfi)->gdi;
+
+		REMMINA_PLUGIN_DEBUG("Unmap event received, enabling TS_SUPPRESS_OUTPUT_PDU ");
+		gdi_send_suppress_output(gdi, TRUE);
+	}
 
 	return FALSE;
 }
@@ -82,6 +104,7 @@ static gboolean remmina_rdp_event_on_unmap(GtkWindow *window, GdkEvent *event, R
 static gboolean remmina_rdp_event_on_focus_in(GtkWidget *widget, GdkEventKey *event, RemminaProtocolWidget *gp)
 {
 	TRACE_CALL(__func__);
+
 	rfContext *rfi = GET_PLUGIN_DATA(gp);
 	rdpInput *input;
 	GdkModifierType state;
@@ -93,10 +116,13 @@ static gboolean remmina_rdp_event_on_focus_in(GtkWidget *widget, GdkEventKey *ev
 #endif
 	GdkDevice *keyboard = NULL;
 
+	const gchar *wname = gtk_widget_get_name(gtk_widget_get_toplevel(widget));
+	REMMINA_PLUGIN_DEBUG("Top level name is: %s", wname);
+
 	if (!rfi || !rfi->connected || rfi->is_reconnecting)
 		return FALSE;
 
-	input = rfi->instance->input;
+	input = rfi->clientContext.context.input;
 	UINT32 toggle_keys_state = 0;
 
 #if GTK_CHECK_VERSION(3, 20, 0)
@@ -133,7 +159,11 @@ void remmina_rdp_event_event_push(RemminaProtocolWidget *gp, const RemminaPlugin
 		return;
 
 	if (rfi->event_queue) {
+#if GLIB_CHECK_VERSION(2,67,3)
+		event = g_memdup2(e, sizeof(RemminaPluginRdpEvent));
+#else
 		event = g_memdup(e, sizeof(RemminaPluginRdpEvent));
+#endif
 		g_async_queue_push(rfi->event_queue, event);
 
 		if (write(rfi->event_pipe[1], "\0", 1)) {
@@ -153,8 +183,8 @@ static void remmina_rdp_event_release_all_keys(RemminaProtocolWidget *gp)
 		rdp_event = g_array_index(rfi->pressed_keys, RemminaPluginRdpEvent, i);
 		if ((rdp_event.type == REMMINA_RDP_EVENT_TYPE_SCANCODE ||
 		     rdp_event.type == REMMINA_RDP_EVENT_TYPE_SCANCODE_UNICODE) &&
-		    rdp_event.key_event.up == False) {
-			rdp_event.key_event.up = True;
+		    rdp_event.key_event.up == false) {
+			rdp_event.key_event.up = true;
 			remmina_rdp_event_event_push(gp, &rdp_event);
 		}
 	}
@@ -180,7 +210,8 @@ static void remmina_rdp_event_release_key(RemminaProtocolWidget *gp, RemminaPlug
 
 			if (rdp_event_2.key_event.key_code == rdp_event.key_event.key_code &&
 			    rdp_event_2.key_event.unicode_code == rdp_event.key_event.unicode_code &&
-			    rdp_event_2.key_event.extended == rdp_event.key_event.extended) {
+			    rdp_event_2.key_event.extended == rdp_event.key_event.extended &&
+				rdp_event_2.key_event.extended1 == rdp_event.key_event.extended1) {
 				g_array_remove_index_fast(rfi->pressed_keys, i);
 				break;
 			}
@@ -192,6 +223,7 @@ static void keypress_list_add(RemminaProtocolWidget *gp, RemminaPluginRdpEvent r
 {
 	TRACE_CALL(__func__);
 	rfContext *rfi = GET_PLUGIN_DATA(gp);
+
 	if (!rdp_event.key_event.key_code)
 		return;
 
@@ -199,7 +231,6 @@ static void keypress_list_add(RemminaProtocolWidget *gp, RemminaPluginRdpEvent r
 		remmina_rdp_event_release_key(gp, rdp_event);
 	else
 		g_array_append_val(rfi->pressed_keys, rdp_event);
-
 }
 
 
@@ -348,7 +379,9 @@ static gboolean remmina_rdp_event_on_draw(GtkWidget *widget, cairo_t *context, R
 		if (rfi->scale == REMMINA_PROTOCOL_WIDGET_SCALE_MODE_SCALED)
 			cairo_scale(context, rfi->scale_x, rfi->scale_y);
 
+		cairo_surface_flush(rfi->surface);
 		cairo_set_source_surface(context, rfi->surface, 0, 0);
+		cairo_surface_mark_dirty(rfi->surface);
 
 		cairo_set_operator(context, CAIRO_OPERATOR_SOURCE);     // Ignore alpha channel from FreeRDP
 		cairo_paint(context);
@@ -382,6 +415,7 @@ static gboolean remmina_rdp_event_delayed_monitor_layout(RemminaProtocolWidget *
 	gchar *monitorids = NULL;
 	guint32 maxwidth = 0;
 	guint32 maxheight = 0;
+
 	remmina_rdp_monitor_get(rfi, &monitorids, &maxwidth, &maxheight);
 
 	REMMINA_PLUGIN_DEBUG("Sending preconfigured monitor layout");
@@ -403,9 +437,9 @@ static gboolean remmina_rdp_event_delayed_monitor_layout(RemminaProtocolWidget *
 			}
 			rdp_event.type = REMMINA_RDP_EVENT_TYPE_SEND_MONITOR_LAYOUT;
 			if (remmina_plugin_service->file_get_int(remminafile, "multimon", FALSE)) {
-				const rdpMonitor* base = freerdp_settings_get_pointer(rfi->settings, FreeRDP_MonitorDefArray);
-				for (gint i = 0; i < freerdp_settings_get_uint32(rfi->settings, FreeRDP_MonitorCount); ++i) {
-					const rdpMonitor* current = &base[i];
+				const rdpMonitor *base = freerdp_settings_get_pointer(rfi->clientContext.context.settings, FreeRDP_MonitorDefArray);
+				for (gint i = 0; i < freerdp_settings_get_uint32(rfi->clientContext.context.settings, FreeRDP_MonitorCount); ++i) {
+					const rdpMonitor *current = &base[i];
 					REMMINA_PLUGIN_DEBUG("Sending display layout n° %d", i);
 					rdp_event.monitor_layout.Flags = current->is_primary;
 					REMMINA_PLUGIN_DEBUG("EVNT MON LAYOUT - Flags: %i", rdp_event.monitor_layout.Flags);
@@ -432,7 +466,6 @@ static gboolean remmina_rdp_event_delayed_monitor_layout(RemminaProtocolWidget *
 				}
 				remmina_rdp_event_event_push(gp, &rdp_event);
 			} else {
-
 				rdp_event.monitor_layout.width = gpwidth;
 				rdp_event.monitor_layout.height = gpheight;
 				rdp_event.monitor_layout.desktopOrientation = desktopOrientation;
@@ -452,6 +485,7 @@ void remmina_rdp_event_send_delayed_monitor_layout(RemminaProtocolWidget *gp)
 {
 	TRACE_CALL(__func__);
 	rfContext *rfi = GET_PLUGIN_DATA(gp);
+
 	if (!rfi || !rfi->connected || rfi->is_reconnecting)
 		return;
 	if (rfi->delayed_monitor_layout_handler) {
@@ -460,7 +494,6 @@ void remmina_rdp_event_send_delayed_monitor_layout(RemminaProtocolWidget *gp)
 	}
 	if (rfi->scale == REMMINA_PROTOCOL_WIDGET_SCALE_MODE_DYNRES)
 		rfi->delayed_monitor_layout_handler = g_timeout_add(500, (GSourceFunc)remmina_rdp_event_delayed_monitor_layout, gp);
-
 }
 
 static gboolean remmina_rdp_event_on_configure(GtkWidget *widget, GdkEventConfigure *event, RemminaProtocolWidget *gp)
@@ -469,6 +502,7 @@ static gboolean remmina_rdp_event_on_configure(GtkWidget *widget, GdkEventConfig
 	/* Called when gp changes its size or position */
 
 	rfContext *rfi = GET_PLUGIN_DATA(gp);
+
 	if (!rfi || !rfi->connected || rfi->is_reconnecting)
 		return FALSE;
 
@@ -525,16 +559,76 @@ static void remmina_rdp_event_reverse_translate_pos_reverse(RemminaProtocolWidge
 	}
 }
 
+void remmina_rdp_mouse_jitter(RemminaProtocolWidget *gp){
+	TRACE_CALL(__func__);
+	RemminaPluginRdpEvent rdp_event = { 0 };
+	RemminaFile *remminafile;
+	rfContext *rfi = GET_PLUGIN_DATA(gp);
+
+	remminafile = remmina_plugin_service->protocol_plugin_get_file(gp);
+	if (remmina_plugin_service->file_get_int(remminafile, "viewonly", FALSE))
+		return;
+
+	rdp_event.type = REMMINA_RDP_EVENT_TYPE_MOUSE;
+	rdp_event.mouse_event.flags = PTR_FLAGS_MOVE;
+	rdp_event.mouse_event.extended = FALSE;
+	rdp_event.mouse_event.x = rfi->last_x;
+	rdp_event.mouse_event.y = rfi->last_y;
+	remmina_rdp_event_event_push(gp, &rdp_event);
+}
+
+//Keys hex values from gdkkeysyms.h
+void remmina_rdp_idle_keypress(RemminaProtocolWidget *gp, int *keypress_opts){
+	TRACE_CALL(__func__);
+	guint keys[2] = { 0, 0 };
+	RemminaFile *remminafile;
+	rfContext *rfi = GET_PLUGIN_DATA(gp);
+	
+	remminafile = remmina_plugin_service->protocol_plugin_get_file(gp);
+	if (remmina_plugin_service->file_get_int(remminafile, "viewonly", FALSE))
+		return;
+	
+	if (*keypress_opts == 0)
+		return;
+
+	switch (*keypress_opts) {
+		case 1:
+			keys[0] = 0xffe9; // Alt_L 
+			keys[1] = 0xff09; // Tab
+			break;
+		
+		case 2:
+			keys[0] = 0xffeb; // Option key
+			keys[1] = 0xff09; // Tab
+			break;
+		default:
+			return;
+	}
+	remmina_plugin_service->protocol_plugin_send_keys_signals(rfi->drawing_area,
+								  keys, G_N_ELEMENTS(keys), GDK_KEY_PRESS | GDK_KEY_RELEASE);
+}
+
 static gboolean remmina_rdp_event_on_motion(GtkWidget *widget, GdkEventMotion *event, RemminaProtocolWidget *gp)
 {
 	TRACE_CALL(__func__);
 	RemminaPluginRdpEvent rdp_event = { 0 };
+	RemminaFile *remminafile;
+	rfContext *rfi = GET_PLUGIN_DATA(gp);
+
+	remminafile = remmina_plugin_service->protocol_plugin_get_file(gp);
+	if (remmina_plugin_service->file_get_int(remminafile, "viewonly", FALSE))
+		return FALSE;
 
 	rdp_event.type = REMMINA_RDP_EVENT_TYPE_MOUSE;
 	rdp_event.mouse_event.flags = PTR_FLAGS_MOVE;
 	rdp_event.mouse_event.extended = FALSE;
 
 	remmina_rdp_event_translate_pos(gp, event->x, event->y, &rdp_event.mouse_event.x, &rdp_event.mouse_event.y);
+	if (rfi != NULL){
+		rfi->last_x = rdp_event.mouse_event.x;
+		rfi->last_y = rdp_event.mouse_event.y;
+	}
+	
 	remmina_rdp_event_event_push(gp, &rdp_event);
 
 	return TRUE;
@@ -551,6 +645,8 @@ static gboolean remmina_rdp_event_on_button(GtkWidget *widget, GdkEventButton *e
 	RemminaFile *remminafile;
 
 	remminafile = remmina_plugin_service->protocol_plugin_get_file(gp);
+	if (remmina_plugin_service->file_get_int(remminafile, "viewonly", FALSE))
+		return FALSE;
 
 	/* We bypass 2button-press and 3button-press events */
 	if ((event->type != GDK_BUTTON_PRESS) && (event->type != GDK_BUTTON_RELEASE))
@@ -615,6 +711,11 @@ static gboolean remmina_rdp_event_on_scroll(GtkWidget *widget, GdkEventScroll *e
 	gint flag;
 	RemminaPluginRdpEvent rdp_event = { 0 };
 	float windows_delta;
+	RemminaFile *remminafile;
+
+	remminafile = remmina_plugin_service->protocol_plugin_get_file(gp);
+	if (remmina_plugin_service->file_get_int(remminafile, "viewonly", FALSE))
+		return FALSE;
 
 	flag = 0;
 	rdp_event.type = REMMINA_RDP_EVENT_TYPE_MOUSE;
@@ -630,20 +731,33 @@ static gboolean remmina_rdp_event_on_scroll(GtkWidget *widget, GdkEventScroll *e
 		flag = PTR_FLAGS_WHEEL | 0x0188;  // -120 (one scroll unit) in 9 bits two's complement
 		break;
 
+	case GDK_SCROLL_RIGHT:
+		flag = PTR_FLAGS_HWHEEL | 0x0078; // 120 is one scroll unit defined in WM_MOUSEWHEEL
+		break;
+
+	case GDK_SCROLL_LEFT:
+		flag = PTR_FLAGS_HWHEEL | 0x0188; // -120 (one scroll unit) in 9 bits two's complement
+		break;
+
 #if GTK_CHECK_VERSION(3, 4, 0)
 	case GDK_SCROLL_SMOOTH:
 
-		if (event->delta_y == 0.0)
+		if (event->delta_y != 0.0) {
+			flag = PTR_FLAGS_WHEEL;
+			windows_delta = event->delta_y * -120;
+		} else if (event->delta_x != 0.0) {
+			flag = PTR_FLAGS_HWHEEL;
+			windows_delta = event->delta_x * 120;
+		} else {
 			return FALSE;
-
-		windows_delta = event->delta_y * -120;
+		}
 
 		if (windows_delta > 255)
 			windows_delta = 255;
 		if (windows_delta < -256)
 			windows_delta = -256;
 
-		flag = PTR_FLAGS_WHEEL | ((short)windows_delta & WheelRotationMask);
+		flag |= ((short)windows_delta & WheelRotationMask);
 
 		break;
 #endif
@@ -702,12 +816,17 @@ static gboolean remmina_rdp_event_on_key(GtkWidget *widget, GdkEventKey *event,
 	rfContext *rfi = GET_PLUGIN_DATA(gp);
 	RemminaPluginRdpEvent rdp_event;
 	RemminaPluginRdpKeymapEntry *kep;
+	RemminaFile *remminafile;
 	DWORD scancode = 0;
 	int ik;
 
 	if (!rfi || !rfi->connected || rfi->is_reconnecting)
 		return FALSE;
 
+	remminafile = remmina_plugin_service->protocol_plugin_get_file(gp);
+	if (remmina_plugin_service->file_get_int(remminafile, "viewonly", FALSE))
+		return FALSE;
+
 #ifdef ENABLE_GTK_INSPECTOR_KEY
 	/* GTK inspector key is propagated up. Disabled by default.
 	 * enable it by defining ENABLE_GTK_INSPECTOR_KEY */
@@ -717,8 +836,9 @@ static gboolean remmina_rdp_event_on_key(GtkWidget *widget, GdkEventKey *event,
 #endif
 
 	rdp_event.type = REMMINA_RDP_EVENT_TYPE_SCANCODE;
-	rdp_event.key_event.up = (event->type == GDK_KEY_PRESS ? False : True);
-	rdp_event.key_event.extended = False;
+	rdp_event.key_event.up = (event->type == GDK_KEY_PRESS ? false : true);
+	rdp_event.key_event.extended = false;
+	rdp_event.key_event.extended1 = false;
 
 	switch (event->keyval) {
 	case GDK_KEY_Pause:
@@ -728,16 +848,20 @@ static gboolean remmina_rdp_event_on_key(GtkWidget *widget, GdkEventKey *event,
 		 * for pause key management
 		 */
 		rdp_event.key_event.key_code = 0x1D;
-		rdp_event.key_event.up = False;
+		rdp_event.key_event.up = false;
+		rdp_event.key_event.extended1 = TRUE;
 		remmina_rdp_event_event_push(gp, &rdp_event);
 		rdp_event.key_event.key_code = 0x45;
-		rdp_event.key_event.up = False;
+		rdp_event.key_event.up = false;
+		rdp_event.key_event.extended1 = FALSE;
 		remmina_rdp_event_event_push(gp, &rdp_event);
 		rdp_event.key_event.key_code = 0x1D;
-		rdp_event.key_event.up = True;
+		rdp_event.key_event.up = true;
+		rdp_event.key_event.extended1 = TRUE;
 		remmina_rdp_event_event_push(gp, &rdp_event);
 		rdp_event.key_event.key_code = 0x45;
-		rdp_event.key_event.up = True;
+		rdp_event.key_event.up = true;
+		rdp_event.key_event.extended1 = FALSE;
 		remmina_rdp_event_event_push(gp, &rdp_event);
 		break;
 
@@ -753,14 +877,30 @@ static gboolean remmina_rdp_event_on_key(GtkWidget *widget, GdkEventKey *event,
 					}
 				}
 			}
+#if FREERDP_VERSION_MAJOR >= 3
+			guint32 keyboard_type = freerdp_settings_get_uint32(rfi->clientContext.context.settings, FreeRDP_KeyboardType);
+			if (keyboard_type == 0){
+				keyboard_type = WINPR_KBD_TYPE_IBM_ENHANCED;
+			}
+#ifdef GDK_WINDOWING_X11
+			DWORD vc = GetVirtualKeyCodeFromKeycode(hardware_keycode, WINPR_KEYCODE_TYPE_XKB);
+#else
+			DWORD vc = GetVirtualKeyCodeFromKeycode(hardware_keycode, WINPR_KEYCODE_TYPE_EVDEV);
+#endif
+			const DWORD sc = GetVirtualScanCodeFromVirtualKeyCode(vc, keyboard_type); 
+			DWORD scancode = freerdp_keyboard_remap_key(rfi->remap_table, sc);
+#else
 			scancode = freerdp_keyboard_get_rdp_scancode_from_x11_keycode(hardware_keycode);
+#endif
 			if (scancode) {
 				rdp_event.key_event.key_code = scancode & 0xFF;
 				rdp_event.key_event.extended = scancode & 0x100;
+				rdp_event.key_event.extended1 = FALSE;
 				remmina_rdp_event_event_push(gp, &rdp_event);
 				keypress_list_add(gp, rdp_event);
 			}
 		} else {
+			hardware_keycode = event->hardware_keycode;
 			unicode_keyval = gdk_keyval_to_unicode(event->keyval);
 			/* Decide when whe should send a keycode or a Unicode character.
 			 * - All non char keys (Shift, Alt, Super) should be sent as keycode
@@ -775,9 +915,25 @@ static gboolean remmina_rdp_event_on_key(GtkWidget *widget, GdkEventKey *event,
 			    unicode_keyval == 0 ||                                                      // Impossible to translate
 			    (event->state & (GDK_MOD1_MASK | GDK_CONTROL_MASK | GDK_SUPER_MASK)) != 0   // A modifier not recognized by gdk_keyval_to_unicode()
 			    ) {
+#if FREERDP_VERSION_MAJOR >= 3
+
+				guint32 keyboard_type = freerdp_settings_get_uint32(rfi->clientContext.context.settings, FreeRDP_KeyboardType);
+				if (keyboard_type == 0){
+					keyboard_type = WINPR_KBD_TYPE_IBM_ENHANCED;
+				}
+#ifdef GDK_WINDOWING_X11
+				DWORD vc = GetVirtualKeyCodeFromKeycode(hardware_keycode, WINPR_KEYCODE_TYPE_XKB);
+#else
+				DWORD vc = GetVirtualKeyCodeFromKeycode(hardware_keycode, WINPR_KEYCODE_TYPE_EVDEV);
+#endif
+				const DWORD sc = GetVirtualScanCodeFromVirtualKeyCode(vc, keyboard_type); 
+				DWORD scancode = freerdp_keyboard_remap_key(rfi->remap_table, sc);
+#else
 				scancode = freerdp_keyboard_get_rdp_scancode_from_x11_keycode(event->hardware_keycode);
+#endif
 				rdp_event.key_event.key_code = scancode & 0xFF;
 				rdp_event.key_event.extended = scancode & 0x100;
+				rdp_event.key_event.extended1 = FALSE;
 				if (rdp_event.key_event.key_code) {
 					remmina_rdp_event_event_push(gp, &rdp_event);
 					keypress_list_add(gp, rdp_event);
@@ -785,7 +941,8 @@ static gboolean remmina_rdp_event_on_key(GtkWidget *widget, GdkEventKey *event,
 			} else {
 				rdp_event.type = REMMINA_RDP_EVENT_TYPE_SCANCODE_UNICODE;
 				rdp_event.key_event.unicode_code = unicode_keyval;
-				rdp_event.key_event.extended = False;
+				rdp_event.key_event.extended = false;
+				rdp_event.key_event.extended1 = FALSE;
 				remmina_rdp_event_event_push(gp, &rdp_event);
 				keypress_list_add(gp, rdp_event);
 			}
@@ -802,24 +959,29 @@ gboolean remmina_rdp_event_on_clipboard(GtkClipboard *gtkClipboard, GdkEvent *ev
 	TRACE_CALL(__func__);
 	RemminaPluginRdpEvent rdp_event = { 0 };
 	CLIPRDR_FORMAT_LIST *pFormatList;
+	GObject *new_owner;
 
-	/* Usually "owner-change" is fired when a user pres "COPY" on the client
+	/* Usually "owner-change" is fired when a user presses "COPY" on the client
 	 * OR when this plugin calls gtk_clipboard_set_with_owner()
-	 * after receivina a RDP server format list in remmina_rdp_cliprdr_server_format_list()
+	 * after receiving a RDP server format list in remmina_rdp_cliprdr_server_format_list()
 	 * In the latter case, we must ignore owner change */
 
-	REMMINA_PLUGIN_DEBUG("owner-change event received");
+	REMMINA_PLUGIN_DEBUG("gp=%p: owner-change event received", gp);
 
 	rfContext *rfi = GET_PLUGIN_DATA(gp);
-	if (rfi)
-		remmina_rdp_clipboard_abort_transfer(rfi);
 
-	if (gtk_clipboard_get_owner(gtkClipboard) != (GObject *)gp) {
+	if (rfi)
+		remmina_rdp_clipboard_abort_client_format_data_request(rfi);
+
+	new_owner = gtk_clipboard_get_owner(gtkClipboard);
+	if (new_owner != (GObject *)gp) {
 		/* To do: avoid this when the new owner is another remmina protocol widget of
 		 * the same remmina application */
-		REMMINA_PLUGIN_DEBUG("     new owner is different than me: new=%p me=%p. Sending local clipboard format list to server.",
-				     gtk_clipboard_get_owner(gtkClipboard), (GObject *)gp);
+		REMMINA_PLUGIN_DEBUG("gp=%p owner-change: new owner is different than me: new=%p me=%p",
+				     gp, new_owner, gp);
 
+		REMMINA_PLUGIN_DEBUG("gp=%p owner-change: new owner is not me: Sending local clipboard format list to server.",
+			gp, new_owner, gp);
 		pFormatList = remmina_rdp_cliprdr_get_client_format_list(gp);
 		rdp_event.type = REMMINA_RDP_EVENT_TYPE_CLIPBOARD_SEND_CLIENT_FORMAT_LIST;
 		rdp_event.clipboard_formatlist.pFormatList = pFormatList;
@@ -839,9 +1001,20 @@ void remmina_rdp_event_init(RemminaProtocolWidget *gp)
 	GtkClipboard *clipboard;
 	RemminaFile *remminafile;
 
+	gboolean disable_smooth_scrolling = FALSE;
+
 	if (!rfi) return;
 	remminafile = remmina_plugin_service->protocol_plugin_get_file(gp);
 
+	/* we get first the global preferences */
+	s = remmina_plugin_service->pref_get_value("rdp_disable_smooth_scrolling");
+	disable_smooth_scrolling = (s && s[0] == '1' ? TRUE : FALSE);
+	g_free(s), s = NULL;
+	/* Otherwise we use the connection profile specific setting */
+	disable_smooth_scrolling = remmina_plugin_service->file_get_int(remminafile, "disable-smooth-scrolling", disable_smooth_scrolling);
+
+	REMMINA_PLUGIN_DEBUG("Disable smooth scrolling is set to %d", disable_smooth_scrolling);
+
 	rfi->drawing_area = gtk_drawing_area_new();
 	gtk_widget_show(rfi->drawing_area);
 	gtk_container_add(GTK_CONTAINER(gp), rfi->drawing_area);
@@ -849,23 +1022,26 @@ void remmina_rdp_event_init(RemminaProtocolWidget *gp)
 	gtk_widget_add_events(rfi->drawing_area, GDK_POINTER_MOTION_MASK
 			      | GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK
 			      | GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK
-#if GTK_CHECK_VERSION(3, 4, 0)
-			      | GDK_SMOOTH_SCROLL_MASK
-#endif
 			      | GDK_SCROLL_MASK | GDK_FOCUS_CHANGE_MASK);
+
+	if (!disable_smooth_scrolling) {
+		REMMINA_PLUGIN_DEBUG("Adding GDK_SMOOTH_SCROLL_MASK");
+		gtk_widget_add_events(rfi->drawing_area, GDK_SMOOTH_SCROLL_MASK);
+	}
+
 	gtk_widget_set_can_focus(rfi->drawing_area, TRUE);
 
 	remmina_plugin_service->protocol_plugin_register_hostkey(gp, rfi->drawing_area);
 
 	s = remmina_plugin_service->pref_get_value("rdp_use_client_keymap");
 	rfi->use_client_keymap = (s && s[0] == '1' ? TRUE : FALSE);
-	g_free(s);
+	g_free(s), s = NULL;
 
 	/* Read special keymap from profile file, if exists */
 	remmina_rdp_event_init_keymap(rfi, remmina_plugin_service->pref_get_value("rdp_map_keycode"));
 
 	if (rfi->use_client_keymap && rfi->keymap)
-		fprintf(stderr, "RDP profile error: you cannot define both rdp_map_hardware_keycode and have 'Use client keuboard mapping' enabled\n");
+		fprintf(stderr, "RDP profile error: you cannot define both rdp_map_hardware_keycode and have 'Use client keyboard mapping' enabled\n");
 
 	g_signal_connect(G_OBJECT(rfi->drawing_area), "draw",
 			 G_CALLBACK(remmina_rdp_event_on_draw), gp);
@@ -885,10 +1061,16 @@ void remmina_rdp_event_init(RemminaProtocolWidget *gp)
 			 G_CALLBACK(remmina_rdp_event_on_key), gp);
 	g_signal_connect(G_OBJECT(rfi->drawing_area), "focus-in-event",
 			 G_CALLBACK(remmina_rdp_event_on_focus_in), gp);
-	g_signal_connect(G_OBJECT(gtk_widget_get_toplevel(rfi->drawing_area)), "map-event",
-			 G_CALLBACK(remmina_rdp_event_on_map), gp);
-	g_signal_connect(G_OBJECT(gtk_widget_get_toplevel(rfi->drawing_area)), "unmap-event",
-			 G_CALLBACK(remmina_rdp_event_on_unmap), gp);
+	/** Fixme: This comment
+	 * needed for TS_SUPPRESS_OUTPUT_PDU
+	 * But it works only when we stay in the same window mode, if we switch to
+	 * fullscreen, for instance, the object reference is lost, so we loose these
+	 * events.
+	 */
+	//g_signal_connect(G_OBJECT(gtk_widget_get_toplevel(rfi->drawing_area)), "map-event",
+	// G_CALLBACK(remmina_rdp_event_on_map), gp);
+	//g_signal_connect(G_OBJECT(gtk_widget_get_toplevel(rfi->drawing_area)), "unmap-event",
+	// G_CALLBACK(remmina_rdp_event_on_unmap), gp);
 
 	if (!remmina_plugin_service->file_get_int(remminafile, "disableclipboard", FALSE)) {
 		clipboard = gtk_widget_get_clipboard(rfi->drawing_area, GDK_SELECTION_CLIPBOARD);
@@ -926,7 +1108,7 @@ void remmina_rdp_event_init(RemminaProtocolWidget *gp)
 #endif
 }
 
-void remmina_rdp_event_free_event(RemminaProtocolWidget *gp, RemminaPluginRdpUiObject *obj)
+void remmina_rdp_event_free_event(RemminaPluginRdpUiObject *obj)
 {
 	TRACE_CALL(__func__);
 
@@ -964,8 +1146,9 @@ void remmina_rdp_event_uninit(RemminaProtocolWidget *gp)
 		rfi->ui_handler = 0;
 	}
 	while ((ui = (RemminaPluginRdpUiObject *)g_async_queue_try_pop(rfi->ui_queue)) != NULL)
-		remmina_rdp_event_free_event(gp, ui);
+		remmina_rdp_event_free_event(ui);
 	if (rfi->surface) {
+		cairo_surface_mark_dirty(rfi->surface);
 		cairo_surface_destroy(rfi->surface);
 		rfi->surface = NULL;
 	}
@@ -1006,11 +1189,13 @@ static void remmina_rdp_event_create_cairo_surface(rfContext *rfi)
 		return;
 
 	if (rfi->surface) {
+		cairo_surface_mark_dirty(rfi->surface);
 		cairo_surface_destroy(rfi->surface);
 		rfi->surface = NULL;
 	}
 	stride = cairo_format_stride_for_width(rfi->cairo_format, gdi->width);
 	rfi->surface = cairo_image_surface_create_for_data((unsigned char *)gdi->primary_buffer, rfi->cairo_format, gdi->width, gdi->height, stride);
+	cairo_surface_flush(rfi->surface);
 }
 
 void remmina_rdp_event_update_scale(RemminaProtocolWidget *gp)
@@ -1033,6 +1218,7 @@ void remmina_rdp_event_update_scale(RemminaProtocolWidget *gp)
 	if (rfi->surface && (cairo_image_surface_get_width(rfi->surface) != gdi->width ||
 			     cairo_image_surface_get_height(rfi->surface) != gdi->height)) {
 		/* Destroys and recreate rfi->surface with new width and height */
+		cairo_surface_mark_dirty(rfi->surface);
 		cairo_surface_destroy(rfi->surface);
 		rfi->surface = NULL;
 		remmina_rdp_event_create_cairo_surface(rfi);
@@ -1074,12 +1260,16 @@ static void remmina_rdp_event_connected(RemminaProtocolWidget *gp, RemminaPlugin
 	remmina_rdp_event_update_scale(gp);
 
 	remmina_plugin_service->protocol_plugin_signal_connection_opened(gp);
+	const gchar *host = freerdp_settings_get_string (rfi->clientContext.context.settings, FreeRDP_ServerHostname);
+	// TRANSLATORS: the placeholder may be either an IP/FQDN or a server hostname
+	REMMINA_PLUGIN_AUDIT(_("Connected to %s via RDP"), host);
 }
 
 static void remmina_rdp_event_reconnect_progress(RemminaProtocolWidget *gp, RemminaPluginRdpUiObject *ui)
 {
 	TRACE_CALL(__func__);
 	rfContext *rfi = GET_PLUGIN_DATA(gp);
+
 	gdk_window_invalidate_rect(gtk_widget_get_window(rfi->drawing_area), NULL, TRUE);
 }
 
@@ -1103,7 +1293,9 @@ static BOOL remmina_rdp_event_create_cursor(RemminaProtocolWidget *gp, RemminaPl
 	}
 
 	surface = cairo_image_surface_create_for_data(data, CAIRO_FORMAT_ARGB32, pointer->width, pointer->height, cairo_format_stride_for_width(CAIRO_FORMAT_ARGB32, pointer->width));
+	cairo_surface_flush(surface);
 	pixbuf = gdk_pixbuf_get_from_surface(surface, 0, 0, pointer->width, pointer->height);
+	cairo_surface_mark_dirty(surface);
 	cairo_surface_destroy(surface);
 	free(data);
 	((rfPointer *)ui->cursor.pointer)->cursor = gdk_cursor_new_from_pixbuf(rfi->display, pixbuf, pointer->xPos, pointer->yPos);
@@ -1124,6 +1316,7 @@ static BOOL remmina_rdp_event_set_pointer_position(RemminaProtocolWidget *gp, gi
 	TRACE_CALL(__func__);
 	GdkWindow *w, *nw;
 	gint nx, ny, wx, wy;
+
 #if GTK_CHECK_VERSION(3, 20, 0)
 	GdkSeat *seat;
 #else
@@ -1213,6 +1406,8 @@ static void remmina_rdp_ui_event_destroy_cairo_surface(RemminaProtocolWidget *gp
 {
 	TRACE_CALL(__func__);
 	rfContext *rfi = GET_PLUGIN_DATA(gp);
+
+	cairo_surface_mark_dirty(rfi->surface);
 	cairo_surface_destroy(rfi->surface);
 	rfi->surface = NULL;
 }
@@ -1282,7 +1477,7 @@ static gboolean remmina_rdp_event_process_ui_queue(RemminaProtocolWidget *gp)
 			pthread_cond_signal(&ui->sync_wait_cond);
 			pthread_mutex_unlock(&ui->sync_wait_mutex);
 		} else {
-			remmina_rdp_event_free_event(gp, ui);
+			remmina_rdp_event_free_event(ui);
 		}
 
 		pthread_mutex_unlock(&rfi->ui_queue_mutex);
@@ -1301,7 +1496,7 @@ static void remmina_rdp_event_queue_ui(RemminaProtocolWidget *gp, RemminaPluginR
 	gboolean ui_sync_save;
 	int oldcanceltype;
 
-	if (rfi->thread_cancelled)
+	if (!rfi || rfi->thread_cancelled)
 		return;
 
 	if (remmina_plugin_service->is_main_thread()) {
@@ -1352,10 +1547,11 @@ int remmina_rdp_event_queue_ui_sync_retint(RemminaProtocolWidget *gp, RemminaPlu
 {
 	TRACE_CALL(__func__);
 	int retval;
+
 	ui->sync = TRUE;
 	remmina_rdp_event_queue_ui(gp, ui);
 	retval = ui->retval;
-	remmina_rdp_event_free_event(gp, ui);
+	remmina_rdp_event_free_event(ui);
 	return retval;
 }
 
@@ -1363,9 +1559,10 @@ void *remmina_rdp_event_queue_ui_sync_retptr(RemminaProtocolWidget *gp, RemminaP
 {
 	TRACE_CALL(__func__);
 	void *rp;
+
 	ui->sync = TRUE;
 	remmina_rdp_event_queue_ui(gp, ui);
 	rp = ui->retptr;
-	remmina_rdp_event_free_event(gp, ui);
+	remmina_rdp_event_free_event(ui);
 	return rp;
 }
