@@ -39,6 +39,10 @@
 #include "vnc_plugin.h"
 #include <rfb/rfbclient.h>
 
+#ifdef HAVE_NETINET_TCP_H
+#include <netinet/tcp.h>
+#endif
+
 #define REMMINA_PLUGIN_VNC_FEATURE_PREF_QUALITY            1
 #define REMMINA_PLUGIN_VNC_FEATURE_PREF_VIEWONLY           2
 #define REMMINA_PLUGIN_VNC_FEATURE_PREF_DISABLESERVERINPUT 3
@@ -1160,8 +1164,11 @@ static gboolean remmina_plugin_vnc_main_loop(RemminaProtocolWidget *gp)
 handle_buffered:
 		if (!HandleRFBServerMessage(cl)) {
 			gpdata->running = FALSE;
-			if (gpdata->connected && !remmina_plugin_service->protocol_plugin_is_closed(gp))
+			// TCP_USER_TIMEOUT should handle connection timeout
+			remmina_plugin_service->protocol_plugin_set_error(gp, "VNC connection timed out");
+			if (gpdata->connected && !remmina_plugin_service->protocol_plugin_is_closed(gp)) {
 				remmina_plugin_service->protocol_plugin_signal_connection_closed(gp);
+			}
 			return FALSE;
 		}
 	}
@@ -1177,6 +1184,7 @@ static gboolean remmina_plugin_vnc_main(RemminaProtocolWidget *gp)
 	rfbClient *cl = NULL;
 	gchar *host;
 	gchar *s = NULL;
+	gint optval;
 	
 	remminafile = remmina_plugin_service->protocol_plugin_get_file(gp);
 	gpdata->running = TRUE;
@@ -1289,8 +1297,15 @@ static gboolean remmina_plugin_vnc_main(RemminaProtocolWidget *gp)
 			REMMINA_PLUGIN_DEBUG("cl->destPort: %d", cl->destPort);
 		}
 
-		cl->appData.useRemoteCursor = (
-			remmina_plugin_service->file_get_int(remminafile, "showcursor", FALSE) ? FALSE : TRUE);
+		int showcursor = remmina_plugin_service->file_get_int(remminafile, "showcursor", FALSE);
+
+		cl->appData.useRemoteCursor = (showcursor ? FALSE : TRUE);
+		if (showcursor){
+			GdkCursor* Cursor = gdk_cursor_new(GDK_BLANK_CURSOR);
+			GdkWindow* win = gtk_widget_get_window((gpdata->drawing_area));
+			gdk_window_set_cursor((win),Cursor);
+		}
+
 
 		remmina_plugin_vnc_update_quality(cl, quality);
 		remmina_plugin_vnc_update_colordepth(cl, colordepth);
@@ -1312,7 +1327,31 @@ static gboolean remmina_plugin_vnc_main(RemminaProtocolWidget *gp)
 		}
 
 		if (rfbInitClient(cl, NULL, NULL)) {
-			REMMINA_PLUGIN_DEBUG("Client initialization successfull");
+			if (cl->sock) {
+#ifdef HAVE_NETINET_TCP_H
+				// SO_KEEPALIVE = good connection should not be closed due to inactivity
+				optval = 1;
+				if (setsockopt(cl->sock, SOL_SOCKET, SO_KEEPALIVE, &optval, sizeof(optval)) < 0) {
+					REMMINA_PLUGIN_DEBUG("TCP KeepAlive not set");
+				}
+				else {
+					REMMINA_PLUGIN_DEBUG("TCP KeepAlive enabled");
+				}
+#ifdef TCP_USER_TIMEOUT
+				optval = remmina_plugin_service->file_get_int(remminafile, "vnc_timeout", 60) * 1000;
+				if (setsockopt(cl->sock, IPPROTO_TCP, TCP_USER_TIMEOUT, &optval, sizeof(optval)) < 0) {
+					REMMINA_PLUGIN_DEBUG("TCP_USER_TIMEOUT not set");
+				}
+				else {
+					REMMINA_PLUGIN_DEBUG("TCP_USER_TIMEOUT set to %i seconds", optval/1000);
+				}
+#endif // TCP_USER_TIMEOUT
+#endif // HAVE_NETINET_TCP_H
+			}
+			else {
+				REMMINA_PLUGIN_DEBUG("Unable to set socket TCP settings");
+			}
+			REMMINA_PLUGIN_DEBUG("Client initialization successful");
 			break;
 		} else {
 			REMMINA_PLUGIN_DEBUG("Client initialization failed");
@@ -1356,8 +1395,15 @@ static gboolean remmina_plugin_vnc_main(RemminaProtocolWidget *gp)
 
 	if (!gpdata->connected) {
 		REMMINA_PLUGIN_DEBUG("Client not connected with error: %s", vnc_error);
-		if (cl && !gpdata->auth_called && !(remmina_plugin_service->protocol_plugin_has_error(gp)))
-			remmina_plugin_service->protocol_plugin_set_error(gp, "%s", vnc_error);
+		if (cl && !gpdata->auth_called && !(remmina_plugin_service->protocol_plugin_has_error(gp))){
+			if (remmina_plugin_service->file_get_int(remminafile, "closeonfailure", FALSE)){
+				//Do nothing to close immediately
+			}
+			else{
+				remmina_plugin_service->protocol_plugin_set_error(gp, "%s", vnc_error);
+			}
+		}
+			
 		gpdata->running = FALSE;
 
 		remmina_plugin_service->protocol_plugin_signal_connection_closed(gp);
@@ -1716,7 +1762,7 @@ static gboolean remmina_plugin_vnc_open_connection(RemminaProtocolWidget *gp)
 	raw_server = remmina_plugin_service->file_get_string(remminafile, "server");
 
 	if (raw_server && strstr(raw_server, "unix://") == raw_server) {
-		REMMINA_PLUGIN_AUDIT(_("Connected to %s via VNC"), server);
+		REMMINA_PLUGIN_AUDIT(_("Connected to %s via VNC"), raw_server);
 	} else {
 		remmina_plugin_service->get_server_port(raw_server,
 				VNC_DEFAULT_PORT,
@@ -1915,7 +1961,6 @@ static gboolean remmina_plugin_vnc_on_size_allocate(GtkWidget *widget, GtkAlloca
 		sprintf(str, "DEBUG: %d x %d", alloc->width, alloc->height);
 		TRACE_CALL(str);
 		if (gpdata->client){
-			rfbClient *cl;
 			SendExtDesktopSize(gpdata->client, alloc->width, alloc->height);
 		}
 	}
@@ -1962,6 +2007,7 @@ static void remmina_plugin_vnc_init(RemminaProtocolWidget *gp)
 	TRACE_CALL(__func__);
 	RemminaPluginVncData *gpdata;
 	gint flags;
+	gdouble aspect_ratio;
 
 	gpdata = g_new0(RemminaPluginVncData, 1);
 	g_object_set_data_full(G_OBJECT(gp), "plugin-data", gpdata, g_free);
@@ -1974,7 +2020,19 @@ static void remmina_plugin_vnc_init(RemminaProtocolWidget *gp)
 
 	gpdata->drawing_area = gtk_drawing_area_new();
 	gtk_widget_show(gpdata->drawing_area);
-	gtk_container_add(GTK_CONTAINER(gp), gpdata->drawing_area);
+
+	aspect_ratio = remmina_plugin_service->file_get_double(remminafile, "aspect_ratio", 0);
+	if (aspect_ratio > 0){
+		GtkWidget* aspectframe = gtk_aspect_frame_new(NULL, 0, 0, aspect_ratio, FALSE);
+
+		gtk_frame_set_shadow_type(GTK_FRAME(aspectframe), GTK_SHADOW_NONE);
+		gtk_widget_show(aspectframe);
+		gtk_container_add(GTK_CONTAINER(aspectframe), gpdata->drawing_area);
+		gtk_container_add(GTK_CONTAINER(gp), aspectframe);
+	}
+	else{
+		gtk_container_add(GTK_CONTAINER(gp), gpdata->drawing_area);
+	}
 
 	gtk_widget_add_events(
 		gpdata->drawing_area,
@@ -2047,6 +2105,13 @@ static gchar vnciport_tooltip[] =
 	   "    Remmina, e.g. with x11vnc:\n"
 	   "    x11vnc -display :0 -connect 192.168.1.36:8888");
 
+static gchar aspect_ratio_tooltip[] =
+	N_("Lock the aspect ratio when dynamic resolution is enabled:\n"
+	   "\n"
+	   "  • The aspect ratio should be entered as a decimal number, e.g. 1.777\n"
+	   "  • 16:9 corresponds roughly to 1.7777, 4:3 corresponds roughly to 1.333\n"
+	   "  • The default value of 0 does not enforce any aspect ratio");	   
+
 static gchar vncencodings_tooltip[] =
 	N_("Overriding the pre-set VNC encoding quality:\n"
 	   "\n"
@@ -2055,6 +2120,12 @@ static gchar vncencodings_tooltip[] =
 	   "  • “Good” sets encoding to “tight zrle ultra copyrect hextile zlib corre rre raw”\n"
 	   "  • “Best (slowest)” sets encoding to “copyrect zrle ultra zlib hextile corre rre raw”");
 
+#ifdef TCP_USER_TIMEOUT
+static gchar vnc_timeout_tooltip[] =
+    N_("VNC timeout length in seconds\n"
+	   "  • This timeout controls how long the client will try to contact the server before dropping the connection");
+#endif // TCP_USER_TIMEOUT	   
+
 /* Array of RemminaProtocolSetting for basic settings.
  * Each item is composed by:
  * a) RemminaProtocolSettingType for setting type
@@ -2108,6 +2179,10 @@ static const RemminaProtocolSetting remmina_plugin_vnci_basic_settings[] =
 static const RemminaProtocolSetting remmina_plugin_vnc_advanced_settings[] =
 {
 	{ REMMINA_PROTOCOL_SETTING_TYPE_TEXT,  "encodings",		 N_("Override pre-set VNC encodings"),	        FALSE, NULL, vncencodings_tooltip },
+	{ REMMINA_PROTOCOL_SETTING_TYPE_TEXT,  "aspect_ratio",		 N_("Dynamic resolution enforced aspect ratio"),	        FALSE, NULL, aspect_ratio_tooltip },
+#ifdef TCP_USER_TIMEOUT
+	{ REMMINA_PROTOCOL_SETTING_TYPE_INT,  "vnc_timeout",    N_("TCP_USER_TIMEOUT length (seconds)"),   FALSE, NULL, vnc_timeout_tooltip },
+#endif // TCP_USER_TIMEOUT
 	{ REMMINA_PROTOCOL_SETTING_TYPE_CHECK, "tightencoding", N_("Force tight encoding"),			TRUE, NULL, N_("Enabling this may help when the remote desktop looks scrambled") },
 	{ REMMINA_PROTOCOL_SETTING_TYPE_CHECK, "disablesmoothscrolling", N_("Disable smooth scrolling"),		FALSE, NULL, NULL },
 	{ REMMINA_PROTOCOL_SETTING_TYPE_CHECK, "disablepasswordstoring", N_("Forget passwords after use"),		TRUE,  NULL, NULL },
@@ -2117,6 +2192,7 @@ static const RemminaProtocolSetting remmina_plugin_vnc_advanced_settings[] =
 	{ REMMINA_PROTOCOL_SETTING_TYPE_CHECK, "disableclipboard",	 N_("Turn off clipboard sync"),			TRUE,  NULL, NULL },
 	{ REMMINA_PROTOCOL_SETTING_TYPE_CHECK, "disableencryption",	 N_("Turn off encryption"),			FALSE, NULL, NULL },
 	{ REMMINA_PROTOCOL_SETTING_TYPE_CHECK, "viewonly",		 N_("View only"),				TRUE, NULL, NULL },
+	{ REMMINA_PROTOCOL_SETTING_TYPE_CHECK, "closeonfailure",		 N_("Close on connection failure"),				TRUE, NULL, NULL },
 	{ REMMINA_PROTOCOL_SETTING_TYPE_END,   NULL,			 NULL,						FALSE, NULL, NULL }
 };
 
