@@ -1,7 +1,7 @@
 /*
  * Remmina - The GTK+ Remote Desktop Client
  * Copyright (C) 2009-2010 Vic Lee
- * Copyright (C) 2014-2022 Antenore Gatta, Giovanni Panozzo
+ * Copyright (C) 2014-2023 Antenore Gatta, Giovanni Panozzo
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -85,8 +85,11 @@ GtkDialog* remmina_ext_exec_new(RemminaFile* remminafile, const char *remmina_ex
 		return FALSE;
 
 	cmd = remmina_file_format_properties(remminafile, cmd);
+	if (cmd != NULL && *cmd != 0) {
+		cmd = remmina_utils_get_flatpak_command(cmd);
+	}
 	g_debug("[%s] updated to: %s", remmina_ext_exec_type, cmd);
-	if (*cmd != 0) {
+	if (cmd != NULL && *cmd != 0) {
 
 		pcspinner = g_new(PCon_Spinner, 1);
 		builder = remmina_public_gtk_builder_new_from_resource("/org/remmina/Remmina/src/../data/ui/remmina_spinner.glade");
@@ -125,7 +128,9 @@ GtkDialog* remmina_ext_exec_new(RemminaFile* remminafile, const char *remmina_ex
 			g_error_free(error);
 		}
 		g_strfreev(argv);
+		g_free(cmd);
 		return (pcspinner->dialog);
 	}
+	g_free(cmd);
 	return FALSE;
 }
