@@ -4,6 +4,7 @@
  * Copyright (C) 2014-2015 Antenore Gatta, Fabio Castelli, Giovanni Panozzo
  * Copyright (C) 2016-2022 Antenore Gatta, Giovanni Panozzo
  * Copyright (C) 2022-2023 Antenore Gatta, Giovanni Panozzo, Hiroyuki Tanaka
+ * Copyright (C) 2023-2024 Hiroyuki Tanaka, Sunil Bhat
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -57,6 +58,7 @@
 #include "remmina_pref_dialog.h"
 #include "remmina_widget_pool.h"
 #include "remmina_plugin_manager.h"
+#include "remmina_bug_report.h"
 #include "remmina_log.h"
 #include "remmina_icon.h"
 #include "remmina_main.h"
@@ -80,6 +82,7 @@ enum {
 	FILENAME_COLUMN,
 	LABELS_COLUMN,
 	NOTES_COLUMN,
+	STATUS_COLUMN,
 	N_COLUMNS
 };
 
@@ -99,6 +102,7 @@ static GActionEntry app_actions[] = {
 	{ "mpchange",	 remmina_main_on_action_application_mpchange,	 NULL, NULL, NULL },
 	{ "plugins",	 remmina_main_on_action_application_plugins,	 NULL, NULL, NULL },
 	{ "preferences", remmina_main_on_action_application_preferences, "i",  NULL, NULL },
+	{ "bug_report",  remmina_main_on_action_application_bug_report, NULL, NULL, NULL},
 	{ "dark",	 remmina_main_on_action_application_dark_theme,	 NULL, NULL, NULL },
 	{ "debug",	 remmina_main_on_action_help_debug,		 NULL, NULL, NULL },
 	{ "community",	 remmina_main_on_action_help_community,		 NULL, NULL, NULL },
@@ -112,6 +116,8 @@ static GActionEntry main_actions[] = {
 	{ "connect",  remmina_main_on_action_connection_connect,	NULL, NULL, NULL },
 	{ "copy",     remmina_main_on_action_connection_copy,		NULL, NULL, NULL },
 	{ "delete",   remmina_main_on_action_connection_delete,		NULL, NULL, NULL },
+	{ "delete_multiple", remmina_main_on_action_connection_delete_multiple, NULL, NULL, NULL },
+	{ "connect_multiple", remmina_main_on_action_connection_connect_multiple, NULL, NULL, NULL },
 	{ "edit",     remmina_main_on_action_connection_edit,		NULL, NULL, NULL },
 	{ "exttools", remmina_main_on_action_connection_external_tools, NULL, NULL, NULL },
 	{ "new",      remmina_main_on_action_connection_new,		NULL, NULL, NULL },
@@ -356,10 +362,24 @@ static void remmina_main_load_file_list_callback(RemminaFile *remminafile, gpoin
 	TRACE_CALL(__func__);
 	GtkTreeIter iter;
 	GtkListStore *store;
-
+	gchar* status_icon = "";
 	store = GTK_LIST_STORE(user_data);
 	gchar *datetime;
-
+	if (remmina_pref_get_boolean("status_check")){
+		status_icon = "org.remmina.Remmina-status-grey";
+		if (g_hash_table_contains(remminamain->network_states, remminafile->filename)){
+			gchar* result = (gchar*)g_hash_table_lookup(remminamain->network_states, remminafile->filename);
+			if (result != NULL){
+				if (strncmp("Yes", result, strlen("Yes")) == 0){
+					status_icon = "org.remmina.Remmina-status-green";
+				}
+				else if (strncmp("No", result, strlen("No")) == 0){
+					status_icon = "org.remmina.Remmina-status-red";
+				}
+			}
+		}
+	}
+	
 	datetime = remmina_file_get_datetime(remminafile);
 	gtk_list_store_append(store, &iter);
 	gtk_list_store_set(store, &iter,
@@ -372,6 +392,7 @@ static void remmina_main_load_file_list_callback(RemminaFile *remminafile, gpoin
 			   DATE_COLUMN, datetime,
 			   FILENAME_COLUMN, remmina_file_get_filename(remminafile),
 			   LABELS_COLUMN, remmina_file_get_string(remminafile, "labels"),
+			   STATUS_COLUMN, status_icon,
 			   -1);
 	g_free(datetime);
 }
@@ -486,6 +507,21 @@ static void remmina_main_load_file_tree_callback(RemminaFile *remminafile, gpoin
 	GtkTreeStore *store;
 	gboolean found;
 	gchar *datetime = NULL;
+	gchar* status_icon = "";
+	if (remmina_pref_get_boolean("status_check")){
+		status_icon = "org.remmina.Remmina-status-grey";
+		if (g_hash_table_contains(remminamain->network_states, remminafile->filename)){
+			gchar* result = (gchar*)g_hash_table_lookup(remminamain->network_states, remminafile->filename);
+			if (result != NULL){
+				if (strncmp("Yes", result, strlen("Yes")) == 0){
+					status_icon = "org.remmina.Remmina-status-green";
+				}
+				else if (strncmp("No", result, strlen("No")) == 0){
+					status_icon = "org.remmina.Remmina-status-red";
+				}
+			}
+		}
+	}
 
 	store = GTK_TREE_STORE(user_data);
 
@@ -495,7 +531,6 @@ static void remmina_main_load_file_tree_callback(RemminaFile *remminafile, gpoin
 							 remmina_file_get_string(remminafile, "group"));
 
 	datetime = remmina_file_get_datetime(remminafile);
-	//REMMINA_DEBUG("The date is %s", datetime);
 	gtk_tree_store_append(store, &child, (found ? &iter : NULL));
 	gtk_tree_store_set(store, &child,
 			   PROTOCOL_COLUMN, remmina_file_get_icon_name(remminafile),
@@ -507,6 +542,7 @@ static void remmina_main_load_file_tree_callback(RemminaFile *remminafile, gpoin
 			   DATE_COLUMN, datetime,
 			   FILENAME_COLUMN, remmina_file_get_filename(remminafile),
 			   LABELS_COLUMN, remmina_file_get_string(remminafile, "labels"),
+			   STATUS_COLUMN, status_icon,
 			   -1);
 	g_free(datetime);
 }
@@ -675,7 +711,7 @@ static void remmina_main_load_files()
 	switch (view_file_mode) {
 	case REMMINA_VIEW_FILE_TREE:
 		/* Create new GtkTreeStore model */
-		newmodel = GTK_TREE_MODEL(gtk_tree_store_new(9, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING));
+		newmodel = GTK_TREE_MODEL(gtk_tree_store_new(10, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING));
 		/* Hide the Group column in the tree view mode */
 		gtk_tree_view_column_set_visible(remminamain->column_files_list_group, FALSE);
 		/* Load groups first */
@@ -687,7 +723,7 @@ static void remmina_main_load_files()
 	case REMMINA_VIEW_FILE_LIST:
 	default:
 		/* Create new GtkListStore model */
-		newmodel = GTK_TREE_MODEL(gtk_list_store_new(9, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING));
+		newmodel = GTK_TREE_MODEL(gtk_list_store_new(10, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING));
 		/* Show the Group column in the list view mode */
 		gtk_tree_view_column_set_visible(remminamain->column_files_list_group, TRUE);
 		/* Load files list */
@@ -764,14 +800,38 @@ void remmina_main_load_files_cb(GtkEntry *entry, char *string, gpointer user_dat
 	remmina_main_load_files();
 }
 
+
+static void remmina_main_load_by_group_callback(RemminaFile *remminafile, gpointer user_data)
+{
+	
+	const gchar* group = remmina_file_get_string(remminafile, "group");
+
+	if (g_strcmp0(remminamain->priv->selected_name, group) == 0 ){
+		if (remmina_pref_get_boolean("use_primary_password")
+			&& remmina_pref_get_boolean("lock_connect")
+			&& remmina_unlock_new(remminamain->window) == 0)
+			return;
+		if (remmina_file_get_int (remminafile, "profile-lock", FALSE) == 1
+				&& remmina_unlock_new(remminamain->window) == 0)
+			return;
+
+		remmina_file_touch(remminafile);
+		rcw_open_from_filename(remminafile->filename);
+	}
+}
+
 void remmina_main_on_action_connection_connect(GSimpleAction *action, GVariant *param, gpointer data)
 {
 	TRACE_CALL(__func__);
 
 	RemminaFile *remminafile;
 
-	if (!remminamain->priv->selected_filename)
+	if (!remminamain->priv->selected_filename){
+		if (remminamain->priv->selected_name){
+			remmina_file_manager_iterate((GFunc)remmina_main_load_by_group_callback, NULL);
+		}
 		return;
+	}
 
 	remminafile = remmina_file_load(remminamain->priv->selected_filename);
 
@@ -854,6 +914,7 @@ void remmina_main_on_action_connection_new(GSimpleAction *action, GVariant *para
 		return;
 	GtkWidget *widget;
 
+	remmina_plugin_manager_get_available_plugins();
 	if (remmina_pref_get_boolean("use_primary_password")
 			&& remmina_pref_get_boolean("lock_edit")
 			&& remmina_unlock_new(remminamain->window) == 0)
@@ -1007,6 +1068,142 @@ void remmina_main_on_action_connection_delete(GSimpleAction *action, GVariant *p
 	remmina_main_clear_selection_data();
 }
 
+
+void remmina_main_on_action_connection_connect_multiple(GSimpleAction *action, GVariant *param, gpointer data)
+{
+	TRACE_CALL(__func__);
+	GtkTreeSelection *sel = gtk_tree_view_get_selection(remminamain->tree_files_list);
+	GtkTreeModel *model = gtk_tree_view_get_model(remminamain->tree_files_list);
+	GList *list = gtk_tree_selection_get_selected_rows(sel, &model);
+	gchar *file_to_load = NULL;
+
+
+	while (list) {
+		GtkTreePath *path = list->data;
+		GtkTreeIter iter;
+		
+		if (!gtk_tree_model_get_iter(model, &iter, path)) {
+			GtkWidget *dialog_warning;
+			dialog_warning = gtk_message_dialog_new(remminamain->window, GTK_DIALOG_MODAL, GTK_MESSAGE_WARNING, GTK_BUTTONS_OK, 
+				_("Failed to load files!"));
+			gtk_dialog_run(GTK_DIALOG(dialog_warning));
+			gtk_widget_destroy(dialog_warning);
+			remmina_main_clear_selection_data();
+			return;
+		}
+
+		gtk_tree_model_get(model, &iter, 
+				FILENAME_COLUMN, &file_to_load, -1);
+
+		if (file_to_load == NULL){
+			gtk_tree_model_get(model, &iter, 
+				GROUP_COLUMN, &file_to_load, -1);
+
+			REMMINA_DEBUG("Group column is %s", file_to_load);
+			if (remminamain->priv->selected_name){
+				remmina_file_manager_iterate((GFunc)remmina_main_load_by_group_callback, NULL);
+			}
+		}
+
+		RemminaFile *remminafile = remmina_file_load(file_to_load);
+
+		if (remminafile == NULL){
+			list = g_list_next(list);
+			continue;
+		}
+
+		if (((remmina_pref_get_boolean("lock_edit")
+				&& remmina_pref_get_boolean("use_primary_password"))
+				|| remmina_file_get_int (remminafile, "profile-lock", FALSE))
+			&& remmina_unlock_new(remminamain->window) == 0){
+				list = g_list_next(list);
+				continue;
+			}
+
+
+		if (remmina_file_get_int (remminafile, "profile-lock", FALSE) == 1
+			&& remmina_unlock_new(remminamain->window) == 0){
+				list = g_list_next(list);
+				continue;
+			}
+				
+
+		remmina_file_touch(remminafile);
+		rcw_open_from_filename(file_to_load);
+
+		
+
+		if (remminafile) {
+			remmina_file_free(remminafile);
+			remminafile = NULL;
+		}
+
+		list = g_list_next(list);
+	}
+	
+	remmina_main_clear_selection_data();
+}
+
+
+
+void remmina_main_on_action_connection_delete_multiple(GSimpleAction *action, GVariant *param, gpointer data)
+{
+	TRACE_CALL(__func__);
+	GtkWidget *dialog;
+	GtkTreeSelection *sel = gtk_tree_view_get_selection(remminamain->tree_files_list);
+	GtkTreeModel *model = gtk_tree_view_get_model(remminamain->tree_files_list);
+	GList *list = gtk_tree_selection_get_selected_rows(sel, &model);
+	gchar *file_to_delete;
+
+	dialog = gtk_message_dialog_new(remminamain->window, GTK_DIALOG_MODAL, GTK_MESSAGE_QUESTION, GTK_BUTTONS_YES_NO,
+				_("Are you sure you want to delete the selected files?"));
+
+	// Delete files if Yes is clicked
+	if (gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_YES) {
+		while (list) {
+			GtkTreePath *path = list->data;
+			GtkTreeIter iter;
+			
+			if (!gtk_tree_model_get_iter(model, &iter, path)) {
+				GtkWidget *dialog_warning;
+				dialog_warning = gtk_message_dialog_new(remminamain->window, GTK_DIALOG_MODAL, GTK_MESSAGE_WARNING, GTK_BUTTONS_OK, 
+					_("Failed to delete files!"));
+				gtk_dialog_run(GTK_DIALOG(dialog_warning));
+				gtk_widget_destroy(dialog_warning);
+				gtk_widget_destroy(dialog);
+				remmina_main_clear_selection_data();
+				return;
+			}
+
+			gtk_tree_model_get(model, &iter, 
+					FILENAME_COLUMN, &file_to_delete, -1);
+
+			RemminaFile *remminafile = remmina_file_load(file_to_delete);
+
+			if (((remmina_pref_get_boolean("lock_edit")
+					&& remmina_pref_get_boolean("use_primary_password"))
+					|| remmina_file_get_int (remminafile, "profile-lock", FALSE))
+				&& remmina_unlock_new(remminamain->window) == 0)
+				return;
+
+			if (remminafile) {
+				remmina_file_free(remminafile);
+				remminafile = NULL;
+			}
+
+			gchar *delfilename = g_strdup(file_to_delete);
+			remmina_file_delete(delfilename);
+			g_free(delfilename), delfilename = NULL;
+			remmina_icon_populate_menu();
+			remmina_main_load_files();
+			list = g_list_next(list);
+		}
+	}
+	
+	gtk_widget_destroy(dialog);
+	remmina_main_clear_selection_data();
+}
+
 void remmina_main_on_accel_application_preferences(GSimpleAction *action, GVariant *param, gpointer data)
 {
 	TRACE_CALL(__func__);
@@ -1020,8 +1217,15 @@ void remmina_main_reload_preferences()
 	GtkSettings *settings;
 	settings = gtk_settings_get_default();
 	g_object_set(settings, "gtk-application-prefer-dark-theme", remmina_pref.dark_theme, NULL);
-	remmina_main_on_action_search_toggle(NULL,NULL,NULL);
-	gtk_tree_view_column_set_visible(remminamain->column_files_list_notes, remmina_pref.always_show_notes);
+	if (remminamain) {
+		if(remmina_pref.hide_searchbar){
+			gtk_toggle_button_set_active(remminamain->search_toggle, FALSE);
+		}
+		else{
+			gtk_toggle_button_set_active(remminamain->search_toggle, TRUE);
+		}
+		gtk_tree_view_column_set_visible(remminamain->column_files_list_notes, remmina_pref.always_show_notes);
+	}
 }
 
 void remmina_main_on_action_application_preferences(GSimpleAction *action, GVariant *param, gpointer data)
@@ -1045,9 +1249,7 @@ void remmina_main_on_action_application_preferences(GSimpleAction *action, GVari
 
 	GtkWidget *widget = remmina_pref_dialog_new(tab_num, remminamain->window);
 
-	gtk_widget_show_all(widget);	
-	/* Reload to use new preferences */
-	remmina_main_reload_preferences();
+	gtk_widget_show(widget);	
 }
 
 void remmina_main_on_action_application_default(GSimpleAction *action, GVariant *param, gpointer data)
@@ -1096,6 +1298,24 @@ void remmina_main_on_date_column_sort_clicked()
 	}
 }
 
+void remmina_main_toggle_password_view(GtkWidget *widget, gpointer data)
+{
+	GtkWindow *mainwindow;
+	gboolean visible = gtk_entry_get_visibility(GTK_ENTRY(widget));
+
+	mainwindow = remmina_main_get_window();
+	if (remmina_pref_get_boolean("use_primary_password") && remmina_pref_get_boolean("lock_view_passwords") && remmina_unlock_new(mainwindow) == 0)
+		return;
+
+	if (visible) {
+		gtk_entry_set_visibility(GTK_ENTRY(widget), FALSE);
+		gtk_entry_set_icon_from_icon_name(GTK_ENTRY(widget), GTK_ENTRY_ICON_SECONDARY, "org.remmina.Remmina-password-reveal-symbolic");
+	} else {
+		gtk_entry_set_visibility(GTK_ENTRY(widget), TRUE);
+		gtk_entry_set_icon_from_icon_name(GTK_ENTRY(widget), GTK_ENTRY_ICON_SECONDARY, "org.remmina.Remmina-password-conceal-symbolic");
+	}
+}
+
 static void remmina_main_import_file_list(GSList *files)
 {
 	TRACE_CALL(__func__);
@@ -1139,7 +1359,7 @@ static void remmina_main_import_file_list(GSList *files)
 		remmina_main_load_files();
 }
 
-static void remmina_main_action_tools_import_on_response(GtkDialog *dialog, gint response_id, gpointer user_data)
+static void remmina_main_action_tools_import_on_response(GtkNativeDialog *dialog, gint response_id, gpointer user_data)
 {
 	TRACE_CALL(__func__);
 	GSList *files;
@@ -1148,19 +1368,61 @@ static void remmina_main_action_tools_import_on_response(GtkDialog *dialog, gint
 		files = gtk_file_chooser_get_filenames(GTK_FILE_CHOOSER(dialog));
 		remmina_main_import_file_list(files);
 	}
-	gtk_widget_destroy(GTK_WIDGET(dialog));
+	gtk_native_dialog_destroy(dialog);
+}
+
+static void remmina_set_file_chooser_filters(GtkFileChooser *chooser)
+{
+	GtkFileFilter *filter;
+
+	g_return_if_fail(GTK_IS_FILE_CHOOSER(chooser));
+
+	filter = gtk_file_filter_new();
+	gtk_file_filter_set_name(filter, _("RDP Files"));
+	gtk_file_filter_add_pattern(filter, "*.rdp");
+	gtk_file_filter_add_pattern(filter, "*.rdpx");
+	gtk_file_filter_add_pattern(filter, "*.RDP");
+	gtk_file_filter_add_pattern(filter, "*.RDPX");
+	gtk_file_chooser_add_filter(GTK_FILE_CHOOSER(chooser), filter);
+	gtk_file_chooser_set_filter(GTK_FILE_CHOOSER(chooser), filter);
+
+	filter = gtk_file_filter_new();
+	gtk_file_filter_set_name(filter, _("Virt-Viewer Files"));
+	gtk_file_filter_add_pattern(filter, "*.vv");
+	gtk_file_chooser_add_filter(chooser, filter);
+
+	filter = gtk_file_filter_new();
+	gtk_file_filter_set_name(filter, _("All Files"));
+	gtk_file_filter_add_pattern(filter, "*");
+	gtk_file_chooser_add_filter(chooser, filter);
 }
 
 void remmina_main_on_action_tools_import(GSimpleAction *action, GVariant *param, gpointer data)
 {
 	TRACE_CALL(__func__);
-	GtkWidget *dialog;
+	GtkFileChooserNative *chooser;
 
-	dialog = gtk_file_chooser_dialog_new(_("Import"), remminamain->window, GTK_FILE_CHOOSER_ACTION_OPEN, "Import",
-					     GTK_RESPONSE_ACCEPT, NULL);
-	gtk_file_chooser_set_select_multiple(GTK_FILE_CHOOSER(dialog), TRUE);
-	g_signal_connect(G_OBJECT(dialog), "response", G_CALLBACK(remmina_main_action_tools_import_on_response), NULL);
-	gtk_widget_show(dialog);
+	chooser = gtk_file_chooser_native_new(_("Import"), remminamain->window,
+					      GTK_FILE_CHOOSER_ACTION_OPEN, _("Import"), _("_Cancel"));
+	gtk_native_dialog_set_modal(GTK_NATIVE_DIALOG(chooser), TRUE);
+	remmina_set_file_chooser_filters(GTK_FILE_CHOOSER(chooser));
+	gtk_file_chooser_set_select_multiple(GTK_FILE_CHOOSER(chooser), TRUE);
+	g_signal_connect(chooser, "response", G_CALLBACK(remmina_main_action_tools_import_on_response), NULL);
+	gtk_native_dialog_show(GTK_NATIVE_DIALOG(chooser));
+}
+
+static void on_export_save_response (GtkFileChooserNative *dialog, int response, RemminaFile *remminafile)
+{
+	if (response == GTK_RESPONSE_ACCEPT) {
+		RemminaFilePlugin *plugin = remmina_plugin_manager_get_export_file_handler(remminafile);
+		if (plugin){
+			gchar *path = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(dialog));
+			plugin->export_func(plugin, remminafile, path);
+			g_free(path);
+		}
+	}
+	remmina_file_free(remminafile);
+	gtk_native_dialog_destroy(GTK_NATIVE_DIALOG(dialog));
 }
 
 void remmina_main_on_action_tools_export(GSimpleAction *action, GVariant *param, gpointer data)
@@ -1169,32 +1431,53 @@ void remmina_main_on_action_tools_export(GSimpleAction *action, GVariant *param,
 	RemminaFilePlugin *plugin;
 	RemminaFile *remminafile;
 	GtkWidget *dialog;
+	GtkFileChooserNative *chooser;
+	gchar *export_name;
 
-	if (!remminamain->priv->selected_filename)
+	if (!remminamain->priv->selected_filename) {
+		dialog = gtk_message_dialog_new(remminamain->window, GTK_DIALOG_MODAL, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK,
+						_("Select the connection profile."));
+		g_signal_connect(G_OBJECT(dialog), "response", G_CALLBACK(gtk_widget_destroy), NULL);
+		gtk_widget_show(dialog);
 		return;
+	}
 
 	remminafile = remmina_file_load(remminamain->priv->selected_filename);
-	if (remminafile == NULL)
+	if (remminafile == NULL) {
+		dialog = gtk_message_dialog_new(remminamain->window, GTK_DIALOG_MODAL, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK,
+						_("Remmina couldn't export."));
+		g_signal_connect(G_OBJECT(dialog), "response", G_CALLBACK(gtk_widget_destroy), NULL);
+		gtk_widget_show(dialog);
 		return;
+	}
+
 	plugin = remmina_plugin_manager_get_export_file_handler(remminafile);
 	if (plugin) {
-		dialog = gtk_file_chooser_dialog_new(plugin->export_hints, remminamain->window,
-						     GTK_FILE_CHOOSER_ACTION_SAVE, _("_Save"), GTK_RESPONSE_ACCEPT, NULL);
-		if (gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_ACCEPT)
-			plugin->export_func(plugin, remminafile, gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(dialog)));
-		gtk_widget_destroy(dialog);
-	} else {
+		chooser = gtk_file_chooser_native_new(plugin->export_hints, remminamain->window,
+						      GTK_FILE_CHOOSER_ACTION_SAVE, _("_Save"), _("_Cancel"));
+		gtk_native_dialog_set_modal(GTK_NATIVE_DIALOG(chooser), TRUE);
+		remmina_set_file_chooser_filters(GTK_FILE_CHOOSER(chooser));
+		gtk_file_chooser_set_do_overwrite_confirmation(GTK_FILE_CHOOSER(chooser), TRUE);
+		export_name = g_strdup_printf("%s%s", remminamain->priv->selected_name, plugin->export_ext);
+		gtk_file_chooser_set_current_name(GTK_FILE_CHOOSER(chooser), export_name);
+		g_free(export_name);
+		g_signal_connect(chooser, "response", G_CALLBACK(on_export_save_response), remminafile);
+		gtk_native_dialog_show(GTK_NATIVE_DIALOG(chooser));
+	} else
+	{
+		remmina_file_free(remminafile);
 		dialog = gtk_message_dialog_new(remminamain->window, GTK_DIALOG_MODAL, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK,
 						_("This protocol does not support exporting."));
 		g_signal_connect(G_OBJECT(dialog), "response", G_CALLBACK(gtk_widget_destroy), NULL);
 		gtk_widget_show(dialog);
+		return;
 	}
-	remmina_file_free(remminafile);
 }
 
 void remmina_main_on_action_application_plugins(GSimpleAction *action, GVariant *param, gpointer data)
 {
 	TRACE_CALL(__func__);
+	remmina_plugin_manager_get_available_plugins();
 	remmina_plugin_manager_show(remminamain->window);
 }
 
@@ -1250,6 +1533,12 @@ void remmina_main_on_action_application_about(GSimpleAction *action, GVariant *p
 	remmina_about_open(remminamain->window);
 };
 
+void remmina_main_on_action_application_bug_report(GSimpleAction *action, GVariant *param, gpointer data)
+{
+	TRACE_CALL(__func__);
+	remmina_bug_report_open(remminamain->window);
+};
+
 static gboolean is_empty(const gchar *s)
 {
 	if (s == NULL)
@@ -1281,18 +1570,22 @@ static gboolean remmina_main_quickconnect(void)
 
 	remminafile = remmina_file_new();
 	server = g_strdup(gtk_entry_get_text(remminamain->entry_quick_connect_server));
-	if (g_hostname_to_ascii(server) == NULL)
+	if (g_hostname_to_ascii(server) == NULL) {
+		g_free(server), server = NULL;
 		return FALSE;
+	}
 	/* If server contain /, e.g. vnc://, it won't connect
 	 * We could search for an array of invalid characters, but
 	 * it's better to find a way to correctly parse and validate addresses
 	 */
-	if (g_strrstr(server, "/") != NULL)
+	if (g_strrstr(server, "/") != NULL) {
+		g_free(server), server = NULL;
 		return FALSE;
-	//if (g_str_has_suffix (server, "/"))
-	//return FALSE;
-	if (is_empty(server))
+	}
+	if (is_empty(server)){
+		g_free(server), server = NULL;
 		return FALSE;
+	}
 
 	/* check if server is an IP address and trim whitespace if so */
 	server_trimmed = g_strdup(server);
@@ -1341,7 +1634,6 @@ void remmina_main_on_action_search_toggle(GSimpleAction *action, GVariant *param
 {
 	TRACE_CALL(__func__);
 	REMMINA_DEBUG("Search toggle triggered");
-	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(remminamain->search_toggle), !remmina_pref.hide_searchbar);
 
 	gboolean toggle_status = gtk_toggle_button_get_active(remminamain->search_toggle);
 
@@ -1390,12 +1682,20 @@ gboolean remmina_main_file_list_on_button_press(GtkWidget *widget, GdkEventButto
 {
 	TRACE_CALL(__func__);
 	if (event->button == MOUSE_BUTTON_RIGHT) {
-		if (!kioskmode && kioskmode == FALSE)
+		if (!kioskmode && kioskmode == FALSE) {
 #if GTK_CHECK_VERSION(3, 22, 0)
-			gtk_menu_popup_at_pointer(GTK_MENU(remminamain->menu_popup), (GdkEvent *)event);
+			// For now, if more than one selected row, display only a delete menu option
+			if (gtk_tree_selection_count_selected_rows(gtk_tree_view_get_selection(remminamain->tree_files_list)) > 1) {
+				gtk_menu_popup_at_pointer(GTK_MENU(remminamain->menu_popup_multi), (GdkEvent *)event);
+				return GDK_EVENT_STOP;
+			}
+			else {
+				gtk_menu_popup_at_pointer(GTK_MENU(remminamain->menu_popup), (GdkEvent *)event);
+			}
 #else
 			gtk_menu_popup(remminamain->menu_popup, NULL, NULL, NULL, NULL, event->button, event->time);
 #endif
+		}
 	}
 	return FALSE;
 }
@@ -1526,8 +1826,6 @@ static void remmina_main_init(void)
 	}
 	gtk_combo_box_set_active(GTK_COMBO_BOX(remminamain->combo_quick_connect_protocol), qcp_actidx);
 
-	/* Connect the group accelerators to the GtkWindow */
-	//gtk_window_add_accel_group(remminamain->window, remminamain->accelgroup_shortcuts);
 	/* Set the Quick Connection */
 	gtk_entry_set_activates_default(remminamain->entry_quick_connect_server, TRUE);
 	/* Set the TreeView for the files list */
@@ -1561,6 +1859,15 @@ void remmina_main_on_show(GtkWidget *w, gpointer user_data)
 #endif
 }
 
+void remmina_main_add_network_status(gchar* key, gchar* value)
+{
+	if (remminamain != NULL){
+		g_hash_table_insert(remminamain->network_states, key, value);
+		remmina_main_load_files();
+	}
+	
+}
+
 /* RemminaMain instance */
 GtkWidget *remmina_main_new(void)
 {
@@ -1570,6 +1877,7 @@ GtkWidget *remmina_main_new(void)
 
 	remminamain = g_new0(RemminaMain, 1);
 	remminamain->priv = g_new0(RemminaMainPriv, 1);
+	remminamain->network_states = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, g_free);
 	/* Assign UI widgets to the private members */
 	remminamain->builder = remmina_public_gtk_builder_new_from_resource("/org/remmina/Remmina/src/../data/ui/remmina_main.glade");
 	remminamain->window = GTK_WINDOW(RM_GET_OBJECT("RemminaMain"));
@@ -1594,6 +1902,7 @@ GtkWidget *remmina_main_new(void)
 	remminamain->menu_popup = GTK_MENU(RM_GET_OBJECT("menu_popup"));
 	remminamain->menu_header_button = GTK_MENU_BUTTON(RM_GET_OBJECT("menu_header_button"));
 	remminamain->menu_popup_full = GTK_MENU(RM_GET_OBJECT("menu_popup_full"));
+	remminamain->menu_popup_multi = GTK_MENU(RM_GET_OBJECT("menu_popup_multi"));
 	if (kioskmode && kioskmode == TRUE) {
 		gtk_widget_set_sensitive(GTK_WIDGET(remminamain->menu_popup_full), FALSE);
 		gtk_widget_set_sensitive(GTK_WIDGET(remminamain->menu_header_button), FALSE);
@@ -1683,4 +1992,4 @@ void remmina_main_show_warning_dialog(const gchar *message) {
         gtk_dialog_run(GTK_DIALOG(dialog));
         gtk_widget_destroy(dialog);
     }
-}
\ No newline at end of file
+}
