@@ -2,7 +2,7 @@
  * Remmina - The GTK+ Remote Desktop Client
  * Copyright (C) 2009-2011 Vic Lee
  * Copyright (C) 2014-2015 Antenore Gatta, Fabio Castelli, Giovanni Panozzo
- * Copyright (C) 2016-2021 Antenore Gatta, Giovanni Panozzo
+ * Copyright (C) 2016-2023 Antenore Gatta, Giovanni Panozzo
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -42,8 +42,6 @@
 #include <gmodule.h>
 #include <stdlib.h>
 
-
-
 #include "remmina_chat_window.h"
 #include "remmina_masterthread_exec.h"
 #include "remmina_ext_exec.h"
@@ -71,6 +69,7 @@ struct _RemminaProtocolWidgetPriv {
 
 	gboolean		has_error;
 	gchar *			error_message;
+	gboolean		user_disconnect;
 	/* ssh_tunnels is an array of RemminaSSHTunnel*
 	 * the 1st one is the "main" tunnel, other tunnels are used for example in sftp commands */
 	GPtrArray *		ssh_tunnels;
@@ -89,6 +88,7 @@ struct _RemminaProtocolWidgetPriv {
 	RemminaMessagePanel *	connect_message_panel;
 	RemminaMessagePanel *	listen_message_panel;
 	RemminaMessagePanel *	auth_message_panel;
+	RemminaMessagePanel *	retry_message_panel;
 
 	/* Data saved from the last message_panel when the user confirm */
 	gchar *			username;
@@ -105,7 +105,9 @@ struct _RemminaProtocolWidgetPriv {
 enum panel_type {
 	RPWDT_AUTH,
 	RPWDT_QUESTIONYESNO,
-	RPWDT_AUTHX509
+	RPWDT_AUTHX509,
+	RPWDT_ACCEPT,
+	RPWDT_PROMPT
 };
 
 G_DEFINE_TYPE(RemminaProtocolWidget, remmina_protocol_widget, GTK_TYPE_EVENT_BOX)
@@ -144,8 +146,8 @@ static void remmina_protocol_widget_class_init(RemminaProtocolWidgetClass *klass
 									    G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION, G_STRUCT_OFFSET(RemminaProtocolWidgetClass, update_align), NULL, NULL,
 									    g_cclosure_marshal_VOID__VOID, G_TYPE_NONE, 0);
 	remmina_protocol_widget_signals[LOCK_DYNRES_SIGNAL] = g_signal_new("lock-dynres", G_TYPE_FROM_CLASS(klass),
-									     G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION, G_STRUCT_OFFSET(RemminaProtocolWidgetClass, lock_dynres), NULL, NULL,
-									     g_cclosure_marshal_VOID__VOID, G_TYPE_NONE, 0);
+									   G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION, G_STRUCT_OFFSET(RemminaProtocolWidgetClass, lock_dynres), NULL, NULL,
+									   g_cclosure_marshal_VOID__VOID, G_TYPE_NONE, 0);
 	remmina_protocol_widget_signals[UNLOCK_DYNRES_SIGNAL] = g_signal_new("unlock-dynres", G_TYPE_FROM_CLASS(klass),
 									     G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION, G_STRUCT_OFFSET(RemminaProtocolWidgetClass, unlock_dynres), NULL, NULL,
 									     g_cclosure_marshal_VOID__VOID, G_TYPE_NONE, 0);
@@ -157,16 +159,16 @@ static void remmina_protocol_widget_close_all_tunnels(RemminaProtocolWidget *gp)
 	TRACE_CALL(__func__);
 	int i;
 
-	if (gp->priv->ssh_tunnels) {
-		for(i = 0;i < gp->priv->ssh_tunnels->len; i++) {
+	if (gp->priv && gp->priv->ssh_tunnels) {
+		for (i = 0; i < gp->priv->ssh_tunnels->len; i++) {
 #ifdef HAVE_LIBSSH
 			remmina_ssh_tunnel_free((RemminaSSHTunnel *)gp->priv->ssh_tunnels->pdata[i]);
 #else
-			REMMINA_DEBUG ("LibSSH support turned off, no need to free SSH tunnel data");
+			REMMINA_DEBUG("LibSSH support turned off, no need to free SSH tunnel data");
 #endif
 		}
+		g_ptr_array_set_size(gp->priv->ssh_tunnels, 0);
 	}
-	g_ptr_array_set_size(gp->priv->ssh_tunnels, 0);
 }
 
 
@@ -209,9 +211,10 @@ static void remmina_protocol_widget_destroy(RemminaProtocolWidget *gp, gpointer
 
 	remmina_protocol_widget_close_all_tunnels(gp);
 
-	g_ptr_array_free(gp->priv->ssh_tunnels, TRUE);
-
-	gp->priv->ssh_tunnels = NULL;
+	if (gp->priv && gp->priv->ssh_tunnels) {
+		g_ptr_array_free(gp->priv->ssh_tunnels, TRUE);
+		gp->priv->ssh_tunnels = NULL;
+	}
 }
 
 void remmina_protocol_widget_grab_focus(RemminaProtocolWidget *gp)
@@ -234,6 +237,7 @@ static void remmina_protocol_widget_init(RemminaProtocolWidget *gp)
 
 	priv = g_new0(RemminaProtocolWidgetPriv, 1);
 	gp->priv = priv;
+	gp->priv->user_disconnect = FALSE;
 	gp->priv->closed = TRUE;
 	gp->priv->ssh_tunnels = g_ptr_array_new();
 
@@ -245,6 +249,8 @@ void remmina_protocol_widget_open_connection_real(gpointer data)
 	TRACE_CALL(__func__);
 	RemminaProtocolWidget *gp = REMMINA_PROTOCOL_WIDGET(data);
 
+	REMMINA_DEBUG("Opening connection");
+
 	RemminaProtocolPlugin *plugin;
 	RemminaProtocolFeature *feature;
 	gint num_plugin;
@@ -274,17 +280,26 @@ void remmina_protocol_widget_open_connection_real(gpointer data)
 			feature += num_plugin;
 		}
 #ifdef HAVE_LIBSSH
+		REMMINA_DEBUG("Have SSH");
 		if (num_ssh) {
 			feature->type = REMMINA_PROTOCOL_FEATURE_TYPE_TOOL;
 			feature->id = REMMINA_PROTOCOL_FEATURE_TOOL_SSH;
 			feature->opt1 = _("Connect via SSH from a new terminal");
+			feature->opt1_type_hint = REMMINA_TYPEHINT_STRING;
 			feature->opt2 = "utilities-terminal";
+			feature->opt2_type_hint = REMMINA_TYPEHINT_STRING;
+			feature->opt3 = NULL;
+			feature->opt3_type_hint = REMMINA_TYPEHINT_UNDEFINED;
 			feature++;
 
 			feature->type = REMMINA_PROTOCOL_FEATURE_TYPE_TOOL;
 			feature->id = REMMINA_PROTOCOL_FEATURE_TOOL_SFTP;
 			feature->opt1 = _("Open SFTP transfer…");
+			feature->opt1_type_hint = REMMINA_TYPEHINT_STRING;
 			feature->opt2 = "folder-remote";
+			feature->opt2_type_hint = REMMINA_TYPEHINT_STRING;
+			feature->opt3 = NULL;
+			feature->opt3_type_hint = REMMINA_TYPEHINT_UNDEFINED;
 			feature++;
 		}
 		feature->type = REMMINA_PROTOCOL_FEATURE_TYPE_END;
@@ -318,6 +333,7 @@ void remmina_protocol_widget_open_connection(RemminaProtocolWidget *gp)
 	rco_destroy_message_panel(gp->cnnobj, mp);
 
 	name = remmina_file_get_string(gp->priv->remmina_file, "name");
+	// TRANSLATORS: “%s” is a placeholder for the connection profile name
 	s = g_strdup_printf(_("Connecting to “%s”…"), (name ? name : "*"));
 
 	mp = remmina_message_panel_new();
@@ -329,12 +345,11 @@ void remmina_protocol_widget_open_connection(RemminaProtocolWidget *gp)
 	remmina_protocol_widget_open_connection_real(gp);
 }
 
-static gboolean conn_closed(gpointer data)
-{
+static gboolean conn_closed_real(gpointer data, int button){
 	TRACE_CALL(__func__);
 	RemminaProtocolWidget *gp = (RemminaProtocolWidget *)data;
 
-#ifdef HAVE_LIBSSH
+	#ifdef HAVE_LIBSSH
 	/* This will close all tunnels */
 	remmina_protocol_widget_close_all_tunnels(gp);
 #endif
@@ -343,11 +358,36 @@ static gboolean conn_closed(gpointer data)
 	/* Notify listeners (usually rcw) that the connection is closed */
 	g_signal_emit_by_name(G_OBJECT(gp), "disconnect");
 	return G_SOURCE_REMOVE;
+
+}
+
+static gboolean conn_closed(gpointer data)
+{
+	TRACE_CALL(__func__);
+	RemminaProtocolWidget *gp = (RemminaProtocolWidget *)data;
+	int disconnect_prompt = remmina_file_get_int(gp->priv->remmina_file, "disconnect-prompt", FALSE);
+	if (!gp->priv->user_disconnect && !gp->priv->has_error && disconnect_prompt){
+		const char* msg = "Plugin Disconnected";
+		if (gp->priv->has_error){
+			msg = remmina_protocol_widget_get_error_message(gp);
+			remmina_protocol_widget_set_error(gp, NULL);
+		}
+		gp->priv->user_disconnect = FALSE;
+		RemminaMessagePanel* mp = remmina_message_panel_new();
+		remmina_message_panel_setup_message(mp, msg, (RemminaMessagePanelCallback)conn_closed_real, gp);
+		rco_show_message_panel(gp->cnnobj, mp);
+		return G_SOURCE_REMOVE;
+	}
+	else{
+		return conn_closed_real(gp, 0);
+	}
+
 }
 
 void remmina_protocol_widget_signal_connection_closed(RemminaProtocolWidget *gp)
 {
-	/* Plugin told us that it closed the connection,
+	/* User told us that they closed the connection,
+	 * or the connection was closed with a known error,
 	 * add async event to main thread to complete our close tasks */
 	TRACE_CALL(__func__);
 	gp->priv->closed = TRUE;
@@ -358,13 +398,11 @@ static gboolean conn_opened(gpointer data)
 {
 	TRACE_CALL(__func__);
 	RemminaProtocolWidget *gp = (RemminaProtocolWidget *)data;
-	guint i;
 
 #ifdef HAVE_LIBSSH
 	if (gp->priv->ssh_tunnels) {
-		for(i = 0;i < gp->priv->ssh_tunnels->len; i++) {
+		for (guint i = 0; i < gp->priv->ssh_tunnels->len; i++)
 			remmina_ssh_tunnel_cancel_accept((RemminaSSHTunnel *)gp->priv->ssh_tunnels->pdata[i]);
-		}
 	}
 #endif
 	if (gp->priv->listen_message_panel) {
@@ -375,13 +413,17 @@ static gboolean conn_opened(gpointer data)
 		rco_destroy_message_panel(gp->cnnobj, gp->priv->connect_message_panel);
 		gp->priv->connect_message_panel = NULL;
 	}
+	if (gp->priv->retry_message_panel) {
+		rco_destroy_message_panel(gp->cnnobj, gp->priv->retry_message_panel);
+		gp->priv->retry_message_panel = NULL;
+	}
 	g_signal_emit_by_name(G_OBJECT(gp), "connect");
 	return G_SOURCE_REMOVE;
 }
 
 void remmina_protocol_widget_signal_connection_opened(RemminaProtocolWidget *gp)
 {
-	/* Plugin told us that it closed the connection,
+	/* Plugin told us that it opened the connection,
 	 * add async event to main thread to complete our close tasks */
 	TRACE_CALL(__func__);
 	g_idle_add(conn_opened, (gpointer)gp);
@@ -407,6 +449,7 @@ static gboolean lock_dynres(gpointer data)
 {
 	TRACE_CALL(__func__);
 	RemminaProtocolWidget *gp = (RemminaProtocolWidget *)data;
+
 	g_signal_emit_by_name(G_OBJECT(gp), "lock-dynres");
 	return G_SOURCE_REMOVE;
 }
@@ -415,6 +458,7 @@ static gboolean unlock_dynres(gpointer data)
 {
 	TRACE_CALL(__func__);
 	RemminaProtocolWidget *gp = (RemminaProtocolWidget *)data;
+
 	g_signal_emit_by_name(G_OBJECT(gp), "unlock-dynres");
 	return G_SOURCE_REMOVE;
 }
@@ -437,6 +481,7 @@ static gboolean desktop_resize(gpointer data)
 {
 	TRACE_CALL(__func__);
 	RemminaProtocolWidget *gp = (RemminaProtocolWidget *)data;
+
 	g_signal_emit_by_name(G_OBJECT(gp), "desktop-resize");
 	return G_SOURCE_REMOVE;
 }
@@ -468,9 +513,12 @@ void remmina_protocol_widget_close_connection(RemminaProtocolWidget *gp)
 		/* Connection is already closed by the plugin, but
 		 * rcw is asking to close again (usually after an error panel)
 		 */
+		/* Clear the current error, or "disconnect" signal func will reshow a panel */
+		remmina_protocol_widget_set_error(gp, NULL);
 		g_signal_emit_by_name(G_OBJECT(gp), "disconnect");
 		return;
 	}
+	gp->priv->user_disconnect = TRUE;
 
 	/* Ask the plugin to close, async.
 	 *      The plugin will emit a "disconnect" signal on gp to call our
@@ -480,7 +528,6 @@ void remmina_protocol_widget_close_connection(RemminaProtocolWidget *gp)
 	return;
 }
 
-
 /** Check if the plugin accepts keystrokes.
  */
 gboolean remmina_protocol_widget_plugin_receives_keystrokes(RemminaProtocolWidget *gp)
@@ -502,6 +549,7 @@ void remmina_protocol_widget_send_keystrokes(RemminaProtocolWidget *gp, GtkMenuI
 	guint keyval;
 	GdkKeymapKey *keys;
 	gint n_keys;
+
 	/* Single keystroke replace */
 	typedef struct _KeystrokeReplace {
 		gchar * search;
@@ -518,17 +566,18 @@ void remmina_protocol_widget_send_keystrokes(RemminaProtocolWidget *gp, GtkMenuI
 		{ "\\\\", "\\", GDK_KEY_backslash },
 		{ NULL,	  NULL, 0		  }
 	};
+
 	/* Keystrokes can only be sent to plugins that accepts them */
 	if (remmina_protocol_widget_plugin_receives_keystrokes(gp)) {
 		/* Replace special characters */
 		for (i = 0; keystrokes_replaces[i].replace; i++) {
 			REMMINA_DEBUG("Keystrokes before replacement is \'%s\'", keystrokes);
 			keystrokes = g_strdup(remmina_public_str_replace_in_place(keystrokes,
-							    keystrokes_replaces[i].search,
-							    keystrokes_replaces[i].replace));
+										  keystrokes_replaces[i].search,
+										  keystrokes_replaces[i].replace));
 			REMMINA_DEBUG("Keystrokes after replacement is \'%s\'", keystrokes);
 		}
-		gchar *iter =  g_strdup(keystrokes);
+		gchar *iter = keystrokes;
 		keyvals = (guint *)g_malloc(strlen(keystrokes));
 		while (TRUE) {
 			/* Process each character in the keystrokes */
@@ -574,8 +623,128 @@ void remmina_protocol_widget_send_keystrokes(RemminaProtocolWidget *gp, GtkMenuI
 	return;
 }
 
+/**
+ * Send to the plugin some keystrokes from the content of the clipboard
+ * This is a copy of remmina_protocol_widget_send_keystrokes but it uses the clipboard content
+ * get from remmina_protocol_widget_send_clipboard
+ * Probably we don't need the replacement table.
+ */
+void remmina_protocol_widget_send_clip_strokes(GtkClipboard *clipboard, const gchar *clip_text, gpointer data)
+{
+	TRACE_CALL(__func__);
+	RemminaProtocolWidget *gp = REMMINA_PROTOCOL_WIDGET(data);
+	if (clip_text == NULL){
+		return;
+	}
+	gchar *text = g_utf8_normalize(clip_text, -1, G_NORMALIZE_DEFAULT_COMPOSE);
+	guint *keyvals;
+	gint i;
+	GdkKeymap *keymap = gdk_keymap_get_for_display(gdk_display_get_default());
+	gunichar character;
+	guint keyval;
+	GdkKeymapKey *keys;
+	gint n_keys;
+
+	/* Single keystroke replace */
+	typedef struct _KeystrokeReplace {
+		gchar * search;
+		gchar * replace;
+		guint	keyval;
+	} KeystrokeReplace;
+	/* Special characters to replace */
+	KeystrokeReplace text_replaces[] =
+	{
+		{ "\\n",  "\n", GDK_KEY_Return	  },
+		{ "\\t",  "\t", GDK_KEY_Tab	  },
+		{ "\\b",  "\b", GDK_KEY_BackSpace },
+		{ "\\e",  "\e", GDK_KEY_Escape	  },
+		{ "\\\\", "\\", GDK_KEY_backslash },
+		{ NULL,	  NULL, 0		  }
+	};
+
+	if (remmina_protocol_widget_plugin_receives_keystrokes(gp)) {
+		if (text) {
+			/* Replace special characters */
+			for (i = 0; text_replaces[i].replace; i++) {
+				REMMINA_DEBUG("Text clipboard before replacement is \'%s\'", text);
+				text = g_strdup(remmina_public_str_replace_in_place(text,
+										    text_replaces[i].search,
+										    text_replaces[i].replace));
+				REMMINA_DEBUG("Text clipboard after replacement is \'%s\'", text);
+			}
+			gchar *iter = text;
+			REMMINA_DEBUG("Iter: %s", iter),
+			keyvals = (guint *)g_malloc(strlen(text));
+			while (TRUE) {
+				/* Process each character in the keystrokes */
+				character = g_utf8_get_char_validated(iter, -1);
+				REMMINA_DEBUG("Char: U+%04" G_GINT32_FORMAT"X", character);
+				if (character == 0)
+					break;
+				keyval = gdk_unicode_to_keyval(character);
+				REMMINA_DEBUG("Keyval: %u", keyval);
+				/* Replace all the special character with its keyval */
+				for (i = 0; text_replaces[i].replace; i++) {
+					if (character == text_replaces[i].replace[0]) {
+						keys = g_new0(GdkKeymapKey, 1);
+						keyval = text_replaces[i].keyval;
+						REMMINA_DEBUG("Special Keyval: %u", keyval);
+						/* A special character was generated, no keyval lookup needed */
+						character = 0;
+						break;
+					}
+				}
+				/* Decode character if it’s not a special character */
+				if (character) {
+					/* get keyval without modifications */
+					if (!gdk_keymap_get_entries_for_keyval(keymap, keyval, &keys, &n_keys)) {
+						REMMINA_WARNING("keyval 0x%04x has no keycode!", keyval);
+						iter = g_utf8_find_next_char(iter, NULL);
+						continue;
+					}
+				}
+				/* Add modifier keys */
+				n_keys = 0;
+				if (keys->level & 1)
+					keyvals[n_keys++] = GDK_KEY_Shift_L;
+				if (keys->level & 2)
+					keyvals[n_keys++] = GDK_KEY_Alt_R;
+				/*
+				 * @fixme heap buffer overflow
+				 * In some cases, for example sending \t as the only sequence
+				 * may lead to a buffer overflow
+				 */
+				keyvals[n_keys++] = keyval;
+				/* Send keystroke to the plugin */
+				gp->priv->plugin->send_keystrokes(gp, keyvals, n_keys);
+				g_free(keys);
+				/* Process next character in the keystrokes */
+				iter = g_utf8_find_next_char(iter, NULL);
+			}
+			g_free(keyvals);
+		}
+		g_free(text);
+	}
+	return;
+}
+
+void remmina_protocol_widget_send_clipboard(RemminaProtocolWidget *gp, GObject*widget)
+{
+	TRACE_CALL(__func__);
+	GtkClipboard *clipboard;
+
+	clipboard = gtk_clipboard_get(GDK_SELECTION_CLIPBOARD);
+
+	/* Request the contents of the clipboard, contents_received will be
+	 * called when we do get the contents.
+	 */
+	gtk_clipboard_request_text(clipboard,
+				   remmina_protocol_widget_send_clip_strokes, gp);
+}
+
 gboolean remmina_protocol_widget_plugin_screenshot(RemminaProtocolWidget *gp, RemminaPluginScreenshotData *rpsd)
 {
+	TRACE_CALL(__func__);
 	if (!gp->priv->plugin->get_plugin_screenshot) {
 		REMMINA_DEBUG("plugin screenshot function is not implemented, using core Remmina functionality");
 		return FALSE;
@@ -584,11 +753,35 @@ gboolean remmina_protocol_widget_plugin_screenshot(RemminaProtocolWidget *gp, Re
 	return gp->priv->plugin->get_plugin_screenshot(gp, rpsd);
 }
 
+gboolean remmina_protocol_widget_map_event(RemminaProtocolWidget *gp)
+{
+	TRACE_CALL(__func__);
+	if (!gp->priv->plugin->map_event) {
+		REMMINA_DEBUG("Map plugin function not implemented");
+		return FALSE;
+	}
+
+	REMMINA_DEBUG("Calling plugin mapping function");
+	return gp->priv->plugin->map_event(gp);
+}
+
+gboolean remmina_protocol_widget_unmap_event(RemminaProtocolWidget *gp)
+{
+	TRACE_CALL(__func__);
+	if (!gp->priv->plugin->unmap_event) {
+		REMMINA_DEBUG("Unmap plugin function not implemented");
+		return FALSE;
+	}
+
+	REMMINA_DEBUG("Calling plugin unmapping function");
+	return gp->priv->plugin->unmap_event(gp);
+}
+
 void remmina_protocol_widget_emit_signal(RemminaProtocolWidget *gp, const gchar *signal_name)
 {
 	TRACE_CALL(__func__);
 
-	g_print("Emitting signals should be used from the object itself, not from another object");
+	REMMINA_DEBUG("Emitting signals should be used from the object itself, not from another object");
 	raise(SIGINT);
 
 	if (!remmina_masterthread_exec_is_main_thread()) {
@@ -656,7 +849,7 @@ void remmina_protocol_widget_call_feature_by_ref(RemminaProtocolWidget *gp, cons
 		if (gp->priv->ssh_tunnels && gp->priv->ssh_tunnels->len > 0) {
 			rcw_open_from_file_full(
 				remmina_file_dup_temp_protocol(gp->priv->remmina_file, "SSH"), NULL,
-									(RemminaSSHTunnel *)gp->priv->ssh_tunnels->pdata[0], NULL);
+				(RemminaSSHTunnel *)gp->priv->ssh_tunnels->pdata[0], NULL);
 			return;
 		}
 		break;
@@ -665,7 +858,7 @@ void remmina_protocol_widget_call_feature_by_ref(RemminaProtocolWidget *gp, cons
 		if (gp->priv->ssh_tunnels && gp->priv->ssh_tunnels->len > 0) {
 			rcw_open_from_file_full(
 				remmina_file_dup_temp_protocol(gp->priv->remmina_file, "SFTP"), NULL,
-									gp->priv->ssh_tunnels->pdata[0], NULL);
+				gp->priv->ssh_tunnels->pdata[0], NULL);
 			return;
 		}
 		break;
@@ -753,31 +946,88 @@ static void cancel_init_tunnel_cb(void *cbdata, int btn)
 	printf("Remmina: Cancelling an opening tunnel is not implemented\n");
 }
 
-static RemminaSSHTunnel* remmina_protocol_widget_init_tunnel(RemminaProtocolWidget *gp)
+static RemminaSSHTunnel *remmina_protocol_widget_init_tunnel(RemminaProtocolWidget *gp)
 {
 	TRACE_CALL(__func__);
 	RemminaSSHTunnel *tunnel;
 	gint ret;
 	gchar *msg;
 	RemminaMessagePanel *mp;
+	gboolean partial = FALSE;
+	gboolean cont = FALSE;
 
 	tunnel = remmina_ssh_tunnel_new_from_file(gp->priv->remmina_file);
 
-	REMMINA_DEBUG ("Creating SSH tunnel to “%s” via SSH…", REMMINA_SSH(tunnel)->server);
+	REMMINA_DEBUG("Creating SSH tunnel to “%s” via SSH…", REMMINA_SSH(tunnel)->server);
+	// TRANSLATORS: “%s” is a placeholder for an hostname or an IP address.
 	msg = g_strdup_printf(_("Connecting to “%s” via SSH…"), REMMINA_SSH(tunnel)->server);
 
 	mp = remmina_protocol_widget_mpprogress(gp->cnnobj, msg, cancel_init_tunnel_cb, NULL);
 	g_free(msg);
 
+
+
+	while (1) {
+		if (!partial) {
+			if (!remmina_ssh_init_session(REMMINA_SSH(tunnel))) {
+				REMMINA_DEBUG("SSH Tunnel init session error: %s", REMMINA_SSH(tunnel)->error);
+				remmina_protocol_widget_set_error(gp, REMMINA_SSH(tunnel)->error);
+				// exit the loop here: OK
+				break;
+			}
+		}
+
+		ret = remmina_ssh_auth_gui(REMMINA_SSH(tunnel), gp, gp->priv->remmina_file);
+		REMMINA_DEBUG("Tunnel auth returned %d", ret);
+		switch (ret) {
+		case REMMINA_SSH_AUTH_SUCCESS:
+			REMMINA_DEBUG("Authentication success");
+			break;
+		case REMMINA_SSH_AUTH_PARTIAL:
+			REMMINA_DEBUG("Continue with the next auth method");
+			partial = TRUE;
+			// Continue the loop: OK
+			continue;
+			break;
+		case REMMINA_SSH_AUTH_RECONNECT:
+			REMMINA_DEBUG("Reconnecting…");
+			if (REMMINA_SSH(tunnel)->session) {
+				ssh_disconnect(REMMINA_SSH(tunnel)->session);
+				ssh_free(REMMINA_SSH(tunnel)->session);
+				REMMINA_SSH(tunnel)->session = NULL;
+			}
+			g_free(REMMINA_SSH(tunnel)->callback);
+			// Continue the loop: OK
+			continue;
+			break;
+		case REMMINA_SSH_AUTH_USERCANCEL:
+			REMMINA_DEBUG("Interrupted by the user");
+			// exit the loop here: OK
+			goto BREAK;
+			break;
+		default:
+			REMMINA_DEBUG("Error during the authentication: %s", REMMINA_SSH(tunnel)->error);
+			remmina_protocol_widget_set_error(gp, REMMINA_SSH(tunnel)->error);
+			// exit the loop here: OK
+			goto BREAK;
+		}
+
+
+		cont = TRUE;
+		break;
+	}
+
+#if 0
+
 	if (!remmina_ssh_init_session(REMMINA_SSH(tunnel))) {
-		REMMINA_DEBUG ("Cannot init SSH session with tunnel struct");
+		REMMINA_DEBUG("Cannot init SSH session with tunnel struct");
 		remmina_protocol_widget_set_error(gp, REMMINA_SSH(tunnel)->error);
 		remmina_ssh_tunnel_free(tunnel);
 		return NULL;
 	}
 
 	ret = remmina_ssh_auth_gui(REMMINA_SSH(tunnel), gp, gp->priv->remmina_file);
-	REMMINA_DEBUG ("Tunnel auth returned %d", ret);
+	REMMINA_DEBUG("Tunnel auth returned %d", ret);
 	if (ret != REMMINA_SSH_AUTH_SUCCESS) {
 		if (ret != REMMINA_SSH_AUTH_USERCANCEL)
 			remmina_protocol_widget_set_error(gp, REMMINA_SSH(tunnel)->error);
@@ -785,6 +1035,13 @@ static RemminaSSHTunnel* remmina_protocol_widget_init_tunnel(RemminaProtocolWidg
 		return NULL;
 	}
 
+#endif
+
+BREAK:
+	if (!cont) {
+		remmina_ssh_tunnel_free(tunnel);
+		return NULL;
+	}
 	remmina_protocol_widget_mpdestroy(gp->cnnobj, mp);
 
 	return tunnel;
@@ -797,21 +1054,19 @@ static void cancel_start_direct_tunnel_cb(void *cbdata, int btn)
 {
 	printf("Remmina: Cancelling start_direct_tunnel is not implemented\n");
 }
-#endif
 
 static gboolean remmina_protocol_widget_tunnel_destroy(RemminaSSHTunnel *tunnel, gpointer data)
 {
 	TRACE_CALL(__func__);
 	RemminaProtocolWidget *gp = REMMINA_PROTOCOL_WIDGET(data);
-	guint idx;
-	gboolean found;
+	guint idx = 0;
 
-#if GLIB_CHECK_VERSION(2,54,0)
-	found = g_ptr_array_find(gp->priv->ssh_tunnels, tunnel, &idx);
+#if GLIB_CHECK_VERSION(2, 54, 0)
+	gboolean found = g_ptr_array_find(gp->priv->ssh_tunnels, tunnel, &idx);
 #else
 	int i;
-	found = FALSE;
-	for(i = 0;i < gp->priv->ssh_tunnels->len; i++) {
+	gboolean found = FALSE;
+	for (i = 0; i < gp->priv->ssh_tunnels->len; i++) {
 		if ((RemminaSSHTunnel *)gp->priv->ssh_tunnels->pdata[i] == tunnel) {
 			found = TRUE;
 			idx = i;
@@ -830,6 +1085,7 @@ static gboolean remmina_protocol_widget_tunnel_destroy(RemminaSSHTunnel *tunnel,
 	}
 	return TRUE;
 }
+#endif
 
 /**
  * Start an SSH tunnel if possible and return the host:port string.
@@ -841,9 +1097,9 @@ gchar *remmina_protocol_widget_start_direct_tunnel(RemminaProtocolWidget *gp, gi
 	const gchar *server;
 	const gchar *ssh_tunnel_server;
 	gchar *ssh_tunnel_host, *srv_host, *dest;
-	gint srv_port, ssh_tunnel_port;
+	gint srv_port, ssh_tunnel_port = 0;
 
-	REMMINA_DEBUG ("SSH tunnel initialization…");
+	REMMINA_DEBUG("SSH tunnel initialization…");
 
 	server = remmina_file_get_string(gp->priv->remmina_file, "server");
 	ssh_tunnel_server = remmina_file_get_string(gp->priv->remmina_file, "ssh_tunnel_server");
@@ -851,11 +1107,16 @@ gchar *remmina_protocol_widget_start_direct_tunnel(RemminaProtocolWidget *gp, gi
 	if (!server)
 		return g_strdup("");
 
-	REMMINA_DEBUG ("Calling remmina_public_get_server_port");
+	if (strstr(server, "unix:///") != NULL) {
+		REMMINA_DEBUG("%s is a UNIX socket", server);
+		return g_strdup(server);
+	}
+
+	REMMINA_DEBUG("Calling remmina_public_get_server_port");
 	remmina_public_get_server_port(server, default_port, &srv_host, &srv_port);
-	REMMINA_DEBUG ("Calling remmina_public_get_server_port (tunnel)");
+	REMMINA_DEBUG("Calling remmina_public_get_server_port (tunnel)");
 	remmina_public_get_server_port(ssh_tunnel_server, 22, &ssh_tunnel_host, &ssh_tunnel_port);
-	REMMINA_DEBUG ("server: %s, port: %d", srv_host, srv_port);
+	REMMINA_DEBUG("server: %s, port: %d", srv_host, srv_port);
 
 	if (port_plus && srv_port < 100)
 		/* Protocols like VNC supports using instance number :0, :1, etc. as port number. */
@@ -877,11 +1138,12 @@ gchar *remmina_protocol_widget_start_direct_tunnel(RemminaProtocolWidget *gp, gi
 	if (!tunnel) {
 		g_free(srv_host);
 		g_free(ssh_tunnel_host);
-		REMMINA_DEBUG ("remmina_protocol_widget_init_tunnel failed with error is %s",
-				remmina_protocol_widget_get_error_message(gp));
+		REMMINA_DEBUG("remmina_protocol_widget_init_tunnel failed with error is %s",
+			      remmina_protocol_widget_get_error_message(gp));
 		return NULL;
 	}
 
+	// TRANSLATORS: “%s” is a placeholder for an hostname or an IP address.
 	msg = g_strdup_printf(_("Connecting to “%s” via SSH…"), server);
 	mp = remmina_protocol_widget_mpprogress(gp->cnnobj, msg, cancel_start_direct_tunnel_cb, NULL);
 	g_free(msg);
@@ -893,7 +1155,7 @@ gchar *remmina_protocol_widget_start_direct_tunnel(RemminaProtocolWidget *gp, gi
 		srv_host = g_strdup("127.0.0.1");
 	}
 
-	REMMINA_DEBUG ("Starting tunnel to: %s, port: %d", ssh_tunnel_host, ssh_tunnel_port);
+	REMMINA_DEBUG("Starting tunnel to: %s, port: %d", ssh_tunnel_host, ssh_tunnel_port);
 	if (!remmina_ssh_tunnel_open(tunnel, srv_host, srv_port, remmina_pref.sshtunnel_port)) {
 		g_free(srv_host);
 		g_free(ssh_tunnel_host);
@@ -911,6 +1173,44 @@ gchar *remmina_protocol_widget_start_direct_tunnel(RemminaProtocolWidget *gp, gi
 
 	g_ptr_array_add(gp->priv->ssh_tunnels, tunnel);
 
+
+	//try startup command
+	ssh_channel channel;
+	int rc;
+	const gchar* tunnel_command = remmina_file_get_string(gp->priv->remmina_file, "ssh_tunnel_command");
+	if (tunnel_command != NULL){
+		channel = ssh_channel_new(REMMINA_SSH(tunnel)->session);
+		if (channel == NULL) return g_strdup_printf("127.0.0.1:%i", remmina_pref.sshtunnel_port);
+
+		rc = ssh_channel_open_session(channel);
+		if (rc != SSH_OK)
+		{
+			ssh_channel_free(channel);
+			return g_strdup_printf("127.0.0.1:%i", remmina_pref.sshtunnel_port);
+		}
+		rc = ssh_channel_request_exec(channel, tunnel_command);
+		if (rc != SSH_OK)
+		{
+			ssh_channel_close(channel);
+			ssh_channel_free(channel);
+			return g_strdup_printf("127.0.0.1:%i", remmina_pref.sshtunnel_port);
+		}
+		struct timeval timeout = {10, 0};
+		ssh_channel channels[2];
+		channels[0] = channel;
+		channels[1] = NULL;
+		rc = ssh_channel_select(channels, NULL, NULL, &timeout);
+		if (rc == SSH_OK){
+			char buffer[256];
+			ssh_channel_read(channel, buffer, sizeof(buffer), 0);
+		}
+		
+		REMMINA_DEBUG("Ran startup command");
+		ssh_channel_close(channel);
+		ssh_channel_free(channel);
+	}
+
+
 	return g_strdup_printf("127.0.0.1:%i", remmina_pref.sshtunnel_port);
 
 #else
@@ -945,6 +1245,7 @@ gboolean remmina_protocol_widget_start_reverse_tunnel(RemminaProtocolWidget *gp,
 	if (!(tunnel = remmina_protocol_widget_init_tunnel(gp)))
 		return FALSE;
 
+	// TRANSLATORS: “%i” is a placeholder for a TCP port number.
 	msg = g_strdup_printf(_("Awaiting incoming SSH connection on port %i…"), remmina_file_get_int(gp->priv->remmina_file, "listenport", 0));
 	mp = remmina_protocol_widget_mpprogress(gp->cnnobj, msg, cancel_start_reverse_tunnel_cb, NULL);
 	g_free(msg);
@@ -975,7 +1276,7 @@ gboolean remmina_protocol_widget_ssh_exec(RemminaProtocolWidget *gp, gboolean wa
 	if (gp->priv->ssh_tunnels->len < 1)
 		return FALSE;
 
-	tunnel = (RemminaSSHTunnel*)gp->priv->ssh_tunnels->pdata[0];
+	tunnel = (RemminaSSHTunnel *)gp->priv->ssh_tunnels->pdata[0];
 
 	if ((channel = ssh_channel_new(REMMINA_SSH(tunnel)->session)) == NULL)
 		return FALSE;
@@ -996,10 +1297,12 @@ gboolean remmina_protocol_widget_ssh_exec(RemminaProtocolWidget *gp, gboolean wa
 				ret = TRUE;
 				break;
 			case 127:
+				// TRANSLATORS: “%s” is a place holder for a unix command path.
 				remmina_ssh_set_application_error(REMMINA_SSH(tunnel),
 								  _("The “%s” command is not available on the SSH server."), cmd);
 				break;
 			default:
+				// TRANSLATORS: “%s” is a place holder for a unix command path. “%i” is a placeholder for an error code number.
 				remmina_ssh_set_application_error(REMMINA_SSH(tunnel),
 								  _("Could not run the “%s” command on the SSH server (status = %i)."), cmd, status);
 				break;
@@ -1033,7 +1336,7 @@ static gboolean remmina_protocol_widget_xport_tunnel_init_callback(RemminaSSHTun
 	gint port;
 	gboolean ret;
 
-	REMMINA_DEBUG ("Calling remmina_public_get_server_port");
+	REMMINA_DEBUG("Calling remmina_public_get_server_port");
 	remmina_public_get_server_port(remmina_file_get_string(gp->priv->remmina_file, "server"), 177, &server, &port);
 	ret = ((RemminaXPortTunnelInitFunc)gp->priv->init_func)(gp,
 								tunnel->remotedisplay, (tunnel->bindlocalhost ? "localhost" : server), port);
@@ -1074,10 +1377,11 @@ gboolean remmina_protocol_widget_start_xport_tunnel(RemminaProtocolWidget *gp, R
 	gchar *server;
 	gchar *msg;
 	RemminaMessagePanel *mp;
-	RemminaSSHTunnel* tunnel;
+	RemminaSSHTunnel *tunnel;
 
 	if (!(tunnel = remmina_protocol_widget_init_tunnel(gp))) return FALSE;
 
+	// TRANSLATORS: “%s” is a placeholder for a hostname or IP address.
 	msg = g_strdup_printf(_("Connecting to %s via SSH…"), remmina_file_get_string(gp->priv->remmina_file, "server"));
 	mp = remmina_protocol_widget_mpprogress(gp->cnnobj, msg, cancel_connect_xport_cb, NULL);
 	g_free(msg);
@@ -1088,7 +1392,7 @@ gboolean remmina_protocol_widget_start_xport_tunnel(RemminaProtocolWidget *gp, R
 	tunnel->disconnect_func = remmina_protocol_widget_xport_tunnel_disconnect_callback;
 	tunnel->callback_data = gp;
 
-	REMMINA_DEBUG ("Calling remmina_public_get_server_port");
+	REMMINA_DEBUG("Calling remmina_public_get_server_port");
 	remmina_public_get_server_port(remmina_file_get_string(gp->priv->remmina_file, "server"), 0, &server, NULL);
 	bindlocalhost = (g_strcmp0(REMMINA_SSH(tunnel)->server, server) == 0);
 	g_free(server);
@@ -1114,7 +1418,7 @@ void remmina_protocol_widget_set_display(RemminaProtocolWidget *gp, gint display
 {
 	TRACE_CALL(__func__);
 #ifdef HAVE_LIBSSH
-	RemminaSSHTunnel* tunnel;
+	RemminaSSHTunnel *tunnel;
 	if (gp->priv->ssh_tunnels->len < 1)
 		return;
 	tunnel = (RemminaSSHTunnel *)gp->priv->ssh_tunnels->pdata[0];
@@ -1135,7 +1439,7 @@ gint remmina_protocol_widget_get_multimon(RemminaProtocolWidget *gp)
 	TRACE_CALL(__func__);
 	/* Returns ehenever multi monitor is enabled (1) */
 	gp->priv->multimon = remmina_file_get_int(gp->priv->remmina_file, "multimon", -1);
-	REMMINA_DEBUG ("Multi monitor is set to %d", gp->priv->multimon);
+	REMMINA_DEBUG("Multi monitor is set to %d", gp->priv->multimon);
 	return gp->priv->multimon;
 }
 
@@ -1146,6 +1450,11 @@ gint remmina_protocol_widget_get_profile_remote_height(RemminaProtocolWidget *gp
 	return gp->priv->profile_remote_height;
 }
 
+const gchar* remmina_protocol_widget_get_name(RemminaProtocolWidget *gp)
+{
+	TRACE_CALL(__func__);
+	return gp ? gp->plugin ? gp->plugin->name : NULL : NULL;
+}
 
 gint remmina_protocol_widget_get_width(RemminaProtocolWidget *gp)
 {
@@ -1274,6 +1583,8 @@ static void authpanel_mt_cb(void *user_data, int button)
 			d->gp->priv->cacrl = remmina_message_panel_field_get_filename(d->gp->priv->auth_message_panel, REMMINA_MESSAGE_PANEL_CACRLFILE);
 			d->gp->priv->clientcert = remmina_message_panel_field_get_filename(d->gp->priv->auth_message_panel, REMMINA_MESSAGE_PANEL_CLIENTCERTFILE);
 			d->gp->priv->clientkey = remmina_message_panel_field_get_filename(d->gp->priv->auth_message_panel, REMMINA_MESSAGE_PANEL_CLIENTKEYFILE);
+		} else if (d->dtype == RPWDT_PROMPT) {
+			d->gp->priv->password = remmina_message_panel_field_get_string(d->gp->priv->auth_message_panel, REMMINA_MESSAGE_PANEL_PASSWORD);
 		}
 	}
 
@@ -1299,9 +1610,15 @@ static gboolean remmina_protocol_widget_dialog_mt_setup(gpointer user_data)
 	RemminaMessagePanel *mp;
 	const gchar *s;
 
+	if (d->gp->cnnobj == NULL)
+		return FALSE;
+
 	mp = remmina_message_panel_new();
 
 	if (d->dtype == RPWDT_AUTH) {
+		if (d->pflags & REMMINA_MESSAGE_PANEL_FLAG_USERNAME) {
+			remmina_message_panel_field_set_string(mp, REMMINA_MESSAGE_PANEL_USERNAME, d->default_username);
+		}
 		remmina_message_panel_setup_auth(mp, authpanel_mt_cb, d, d->title, d->strpasswordlabel, d->pflags);
 		remmina_message_panel_field_set_string(mp, REMMINA_MESSAGE_PANEL_USERNAME, d->default_username);
 		if (d->pflags & REMMINA_MESSAGE_PANEL_FLAG_DOMAIN)
@@ -1310,7 +1627,11 @@ static gboolean remmina_protocol_widget_dialog_mt_setup(gpointer user_data)
 		if (d->pflags & REMMINA_MESSAGE_PANEL_FLAG_SAVEPASSWORD)
 			remmina_message_panel_field_set_switch(mp, REMMINA_MESSAGE_PANEL_FLAG_SAVEPASSWORD, (d->default_password == NULL || d->default_password[0] == 0) ? FALSE: TRUE);
 	} else if (d->dtype == RPWDT_QUESTIONYESNO) {
-		remmina_message_panel_setup_question(mp, d->title, authpanel_mt_cb, d);
+		remmina_message_panel_setup_question(mp, d->title, authpanel_mt_cb, d, FALSE);
+	} else if (d->dtype == RPWDT_PROMPT) {
+		remmina_message_panel_setup_prompt(mp, d->title, authpanel_mt_cb, d);
+	} else if (d->dtype == RPWDT_ACCEPT) {
+		remmina_message_panel_setup_question(mp, d->title, authpanel_mt_cb, d, TRUE);
 	} else if (d->dtype == RPWDT_AUTHX509) {
 		remmina_message_panel_setup_auth_x509(mp, authpanel_mt_cb, d);
 		if ((s = remmina_file_get_string(remminafile, "cacert")) != NULL)
@@ -1422,13 +1743,11 @@ static int remmina_protocol_widget_dialog(enum panel_type dtype, RemminaProtocol
 		rcbutton = mpri.response;
 	} else {
 		d->called_from_subthread = TRUE;
-		// pthread_cleanup_push(ptcleanup, (void*)d);
 		pthread_cond_init(&d->pt_cond, NULL);
 		pthread_mutex_init(&d->pt_mutex, NULL);
 		g_idle_add(remmina_protocol_widget_dialog_mt_setup, d);
 		pthread_mutex_lock(&d->pt_mutex);
 		pthread_cond_wait(&d->pt_cond, &d->pt_mutex);
-		// pthread_cleanup_pop(0);
 		pthread_mutex_destroy(&d->pt_mutex);
 		pthread_cond_destroy(&d->pt_cond);
 
@@ -1444,11 +1763,22 @@ static int remmina_protocol_widget_dialog(enum panel_type dtype, RemminaProtocol
 	return rcbutton;
 }
 
+gchar* remmina_protocol_widget_panel_prompt(RemminaProtocolWidget *gp, const char *msg)
+{
+	remmina_protocol_widget_dialog(RPWDT_PROMPT, gp, 0, msg, NULL, NULL, NULL, NULL);
+	return gp->priv->password;
+}
+
 gint remmina_protocol_widget_panel_question_yesno(RemminaProtocolWidget *gp, const char *msg)
 {
 	return remmina_protocol_widget_dialog(RPWDT_QUESTIONYESNO, gp, 0, msg, NULL, NULL, NULL, NULL);
 }
 
+gint remmina_protocol_widget_panel_question_accept(RemminaProtocolWidget *gp, const char *msg)
+{
+	return remmina_protocol_widget_dialog(RPWDT_ACCEPT, gp, 0, msg, NULL, NULL, NULL, NULL);
+}
+
 gint remmina_protocol_widget_panel_auth(RemminaProtocolWidget *gp, RemminaMessagePanelFlags pflags,
 					const gchar *title, const gchar *default_username, const gchar *default_password, const gchar *default_domain, const gchar *password_prompt)
 {
@@ -1534,10 +1864,15 @@ gint remmina_protocol_widget_panel_new_certificate(RemminaProtocolWidget *gp, co
 	// For markup see https://developer.gnome.org/pygtk/stable/pango-markup-language.html
 	s = g_strdup_printf(
 		"<big>%s</big>\n\n%s %s\n%s %s\n%s %s\n\n<big>%s</big>",
+		// TRANSLATORS: The user is asked to verify a new SSL certificate.
 		_("Certificate details:"),
+		// TRANSLATORS: An SSL certificate subject is usually the remote server the user connect to.
 		_("Subject:"), subject,
+		// TRANSLATORS: The name or email of the entity that have issued the SSL certificate
 		_("Issuer:"), issuer,
+		// TRANSLATORS: An SSL certificate fingerprint, is a hash of a certificate calculated on all certificate's data and its signature.
 		_("Fingerprint:"), fingerprint,
+		// TRANSLATORS: The user is asked to accept or refuse a new SSL certificate.
 		_("Accept certificate?"));
 	rc = remmina_protocol_widget_dialog(RPWDT_QUESTIONYESNO, gp, 0, s, NULL, NULL, NULL, NULL);
 	g_free(s);
@@ -1561,11 +1896,17 @@ gint remmina_protocol_widget_panel_changed_certificate(RemminaProtocolWidget *gp
 	// For markup see https://developer.gnome.org/pygtk/stable/pango-markup-language.html
 	s = g_strdup_printf(
 		"<big>%s</big>\n\n%s %s\n%s %s\n%s %s\n%s %s\n\n<big>%s</big>",
+		// TRANSLATORS: The user is asked to verify a new SSL certificate.
 		_("The certificate changed! Details:"),
+		// TRANSLATORS: An SSL certificate subject is usually the remote server the user connect to.
 		_("Subject:"), subject,
+		// TRANSLATORS: The name or email of the entity that have issued the SSL certificate
 		_("Issuer:"), issuer,
+		// TRANSLATORS: An SSL certificate fingerprint, is a hash of a certificate calculated on all certificate's data and its signature.
 		_("Old fingerprint:"), old_fingerprint,
+		// TRANSLATORS: An SSL certificate fingerprint, is a hash of a certificate calculated on all certificate's data and its signature.
 		_("New fingerprint:"), new_fingerprint,
+		// TRANSLATORS: The user is asked to accept or refuse a new SSL certificate.
 		_("Accept changed certificate?"));
 	rc = remmina_protocol_widget_dialog(RPWDT_QUESTIONYESNO, gp, 0, s, NULL, NULL, NULL, NULL);
 	g_free(s);
@@ -1708,6 +2049,7 @@ void remmina_protocol_widget_panel_show_listen(RemminaProtocolWidget *gp, gint p
 
 	mp = remmina_message_panel_new();
 	s = g_strdup_printf(
+		// TRANSLATORS: “%i” is a placeholder for a port number. “%s”  is a placeholder for a protocol name (VNC).
 		_("Listening on port %i for an incoming %s connection…"), port,
 		remmina_file_get_string(gp->priv->remmina_file, "protocol"));
 	remmina_message_panel_setup_progress(mp, s, NULL, NULL);
@@ -1734,6 +2076,7 @@ void remmina_protocol_widget_panel_show_retry(RemminaProtocolWidget *gp)
 
 	mp = remmina_message_panel_new();
 	remmina_message_panel_setup_progress(mp, _("Could not authenticate, attempting reconnection…"), NULL, NULL);
+	gp->priv->retry_message_panel = mp;
 	rco_show_message_panel(gp->cnnobj, mp);
 }
 
@@ -1812,6 +2155,7 @@ void remmina_protocol_widget_setup(RemminaProtocolWidget *gp, RemminaFile *remmi
 									    remmina_file_get_string(remminafile, "protocol"));
 
 	if (!plugin || !plugin->init || !plugin->open_connection) {
+		// TRANSLATORS: “%s” is a placeholder for a protocol name, like “RDP”.
 		remmina_protocol_widget_set_error(gp, _("Install the %s protocol plugin first."),
 						  remmina_file_get_string(remminafile, "protocol"));
 		gp->priv->plugin = NULL;
@@ -1824,11 +2168,13 @@ void remmina_protocol_widget_setup(RemminaProtocolWidget *gp, RemminaFile *remmi
 	gp->priv->scaler_expand = remmina_file_get_int(gp->priv->remmina_file, "scaler_expand", FALSE);
 }
 
-GtkWindow* remmina_protocol_widget_get_gtkwindow(RemminaProtocolWidget *gp) {
+GtkWindow *remmina_protocol_widget_get_gtkwindow(RemminaProtocolWidget *gp)
+{
 	return rcw_get_gtkwindow(gp->cnnobj);
 }
 
-GtkWidget *remmina_protocol_widget_gtkviewport(RemminaProtocolWidget *gp) {
+GtkWidget *remmina_protocol_widget_gtkviewport(RemminaProtocolWidget *gp)
+{
 	return rcw_get_gtkviewport(gp->cnnobj);
 }
 
@@ -1864,6 +2210,7 @@ void remmina_protocol_widget_send_keys_signals(GtkWidget *widget, const guint *k
 			event.keyval = keyvals[i];
 			event.hardware_keycode = remmina_public_get_keycode_for_keyval(keymap, event.keyval);
 			event.is_modifier = (int)remmina_public_get_modifier_for_keycode(keymap, event.hardware_keycode);
+			REMMINA_DEBUG("Sending keyval: %u, hardware_keycode: %u", event.keyval, event.hardware_keycode);
 			g_signal_emit_by_name(G_OBJECT(widget), "key-press-event", &event, &result);
 		}
 	}
@@ -1911,8 +2258,9 @@ void remmina_protocol_widget_update_remote_resolution(RemminaProtocolWidget *gp)
 		/* Use internal window size as remote desktop size */
 		GtkAllocation al;
 		gtk_widget_get_allocation(GTK_WIDGET(gp), &al);
-		w = al.width;
-		h = al.height;
+		/* use a multiple of four to mitigate scaling when remote host rounds up */
+		w = al.width - al.width % 4;
+		h = al.height - al.height % 4;
 		if (w < 10) {
 			printf("Remmina warning: %s RemminaProtocolWidget w=%d h=%d are too small, adjusting to 640x480\n", __func__, w, h);
 			w = 640;
